"""Script to download and extract Prefix data from GS1."""

import json
from typing import Union

import bs4
import httpx

TrieNode = Union[dict[str, "TrieNode"], tuple[int, str]]  # noqa: UP007

PREFIX_URL = "https://www.gs1.org/standards/id-keys/company-prefix"


def main() -> None:
    """The script's main function."""
    html_content = download(PREFIX_URL)
    trie = parse(html_content)
    output(trie)


def download(url: str) -> bytes:
    """Download the data from GS1."""
    return httpx.get(url, timeout=30).content


def parse(html_content: bytes) -> TrieNode:
    """Parse the data from HTML to _GS1PrefixRange objects."""
    trie: TrieNode = {}

    page = bs4.BeautifulSoup(html_content, "html.parser")
    datatable = page.find("table", {"class": ["table"]})
    assert isinstance(datatable, bs4.element.Tag)
    tbody = datatable.find("tbody")
    assert isinstance(tbody, bs4.element.Tag)

    for row in tbody.find_all("tr"):
        columns = row.find_all("td")

        prefixes: str = columns[0].text.strip()
        usage: str = columns[1].text.strip()

        # Standardize separator between ranges
        prefixes = prefixes.replace("  ", "&")

        # Standardize separator between range ends
        prefixes = prefixes.replace("\u2013", "-")

        # Strip note references from end of usage description
        if ", see" in usage:
            note_start = usage.index(", see")
            usage = usage[:note_start]

        for range_ in prefixes.split("&"):
            if "-" in range_:
                start, end = map(str.strip, range_.split("-"))
            else:
                start, end = range_.strip(), range_.strip()

            length = len(start)
            min_value = int(start)
            max_value = int(end)

            # Build/traverse the trie
            for prefix in range(min_value, max_value + 1):
                node: TrieNode = trie
                digits = list(str(prefix).zfill(length))
                while digits:
                    assert isinstance(node, dict)
                    digit = digits.pop(0)
                    if digits:
                        if digit not in node:
                            node[digit] = {}
                        node = node[digit]
                    else:
                        node[digit] = (length, usage)

    # Exceptions defined in the GS1 prefix table's footnotes
    trie["9"]["6"]["0"] = (3, "GS1 UK - GTIN-8")  # type: ignore  # noqa: PGH003
    trie["9"]["6"]["1"] = (3, "GS1 UK - GTIN-8")  # type: ignore  # noqa: PGH003
    trie["9"]["6"]["2"] = {  # type: ignore  # noqa: PGH003
        "0": (4, "GS1 UK - GTIN-8"),
        "1": (4, "GS1 UK - GTIN-8"),
        "2": (4, "GS1 UK - GTIN-8"),
        "3": (4, "GS1 UK - GTIN-8"),
        "4": (4, "GS1 UK - GTIN-8"),
        "5": (4, "GS1 Poland - GTIN-8"),
        "6": (4, "GS1 Poland - GTIN-8"),
        "7": (4, "GS1 Global Office - GTIN-8"),
        "8": (4, "GS1 Global Office - GTIN-8"),
        "9": (4, "GS1 Global Office - GTIN-8"),
    }

    return trie


def output(trie: TrieNode) -> None:
    """Output the trie data structure as JSON to stdout."""
    print(json.dumps(trie, indent=2))


if __name__ == "__main__":
    main()



"""The top-level Biip parser."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, Any

from biip import ParseConfig, ParseError
from biip.gs1_messages import GS1Message
from biip.gs1_web_uris import GS1WebURI
from biip.gtin import Gtin, GtinFormat
from biip.sscc import Sscc
from biip.symbology import GS1Symbology, SymbologyIdentifier
from biip.upc import Upc

if TYPE_CHECKING:
    from collections.abc import Iterator

    from biip.gs1_element_strings import GS1ElementStrings


def parse(
    value: str,
    *,
    config: ParseConfig | None = None,
) -> ParseResult:
    """Identify data format and parse data.

    The current strategy is:

    1. If Symbology Identifier prefix indicates a GTIN or GS1 Message,
       attempt to parse and validate as that.
    2. Else, if not Symbology Identifier, attempt to parse with all parsers.

    Args:
        value: The data to classify and parse.
        config: Configuration options for parsers.

    Returns:
        A `ParseResult` object with the results and errors from all parsers.
    """
    if config is None:
        config = ParseConfig()

    value = value.strip()
    result = ParseResult(value=value)

    # Extract Symbology Identifier
    if value.startswith("]"):
        result = replace(
            result, symbology_identifier=SymbologyIdentifier.extract(value)
        )
        assert result.symbology_identifier
        value = value[len(result.symbology_identifier) :]

    # Select parsers
    queue: ParseQueue = []
    if result.symbology_identifier is not None:
        if result.symbology_identifier.gs1_symbology in GS1Symbology.with_gtin():
            queue.append((_parse_gtin, value))
        if (
            result.symbology_identifier.gs1_symbology
            in GS1Symbology.with_gs1_messages()
        ):
            queue.append((_parse_gs1_message, value))
        if result.symbology_identifier.gs1_symbology in GS1Symbology.with_gs1_web_uri():
            queue.append((_parse_gs1_web_uri, value))
    elif value.startswith("http"):
        queue.append((_parse_gs1_web_uri, value))
    if not queue:
        # If we're not able to select a subset based on Symbology Identifiers,
        # run all parsers on the full value.
        queue = [
            (_parse_gs1_message, value),
            (_parse_gtin, value),
            (_parse_sscc, value),
            (_parse_upc, value),
        ]

    # Work through queue of parsers and the values to run them on. Any parser may
    # add additional work to the queue. Only the first result for a field is kept.
    while queue:
        (parse_func, val) = queue.pop(0)
        result = parse_func(val, config, queue, result)

    return result


@dataclass(frozen=True)
class ParseResult:
    """Results from a successful barcode parsing."""

    value: str
    """The raw value. Only stripped of surrounding whitespace."""

    symbology_identifier: SymbologyIdentifier | None = None
    """The Symbology Identifier, if any."""

    gtin: Gtin | None = None
    """The extracted [GTIN][biip.gtin.Gtin], if any.

    Is also set if a GS1 Message containing a GTIN was successfully parsed."""

    gtin_error: str | None = None
    """The GTIN parse error, if parsing as a GTIN was attempted and failed."""

    upc: Upc | None = None
    """The extracted [UPC][biip.upc.Upc], if any."""

    upc_error: str | None = None
    """The UPC parse error, if parsing as an UPC was attempted and failed."""

    sscc: Sscc | None = None
    """The extracted [SSCC][biip.sscc.Sscc], if any.

    Is also set if a GS1 Message containing an SSCC was successfully parsed.
    """

    sscc_error: str | None = None
    """The SSCC parse error, if parsing as an SSCC was attempted and failed."""

    gs1_message: GS1Message | None = None
    """The extracted [GS1 Message][biip.gs1_messages.GS1Message], if any."""

    gs1_message_error: str | None = None
    """The GS1 Message parse error.

    If parsing as a GS1 Message was attempted and failed.
    """

    gs1_web_uri: GS1WebURI | None = None
    """The extracted [GS1 Web URI][biip.gs1_web_uris.GS1WebURI], if any."""

    gs1_web_uri_error: str | None = None
    """The GS1 Web URI parse error.

    If parsing as a GS1 Web URI was attempted and failed.
    """

    def __rich_repr__(self) -> Iterator[tuple[str, Any] | tuple[str, Any, Any]]:
        # Skip printing fields with default values
        yield "value", self.value
        yield "symbology_identifier", self.symbology_identifier, None
        yield "gtin", self.gtin, None
        yield "gtin_error", self.gtin_error, None
        yield "upc", self.upc, None
        yield "upc_error", self.upc_error, None
        yield "sscc", self.sscc, None
        yield "sscc_error", self.sscc_error, None
        yield "gs1_message", self.gs1_message, None
        yield "gs1_message_error", self.gs1_message_error, None
        yield "gs1_web_uri", self.gs1_web_uri, None
        yield "gs1_web_uri_error", self.gs1_web_uri_error, None


ParseQueue = list[tuple["Parser", str]]
Parser = Callable[[str, ParseConfig, ParseQueue, ParseResult], ParseResult]


def _parse_gtin(
    value: str,
    config: ParseConfig,
    queue: ParseQueue,
    result: ParseResult,
) -> ParseResult:
    if result.gtin is not None:
        return result  # pragma: no cover

    try:
        gtin = Gtin.parse(value, config=config)
        gtin_error = None
    except ParseError as exc:
        gtin = None
        gtin_error = str(exc)

    result = replace(
        result,
        gtin=gtin,
        gtin_error=gtin_error,
    )

    # If GTIN is a GTIN-12, set UPC on the top-level result.
    if result.gtin is not None and result.gtin.format == GtinFormat.GTIN_12:
        queue.append((_parse_upc, result.gtin.as_gtin_12()))

    return result


def _parse_upc(
    value: str,
    config: ParseConfig,
    queue: ParseQueue,
    result: ParseResult,
) -> ParseResult:
    if result.upc is not None:
        return result  # pragma: no cover

    try:
        upc = Upc.parse(value, config=config)
        upc_error = None
    except ParseError as exc:
        upc = None
        upc_error = str(exc)

    if upc is not None:
        # If UPC, expand and set GTIN on the top-level result.
        queue.append((_parse_gtin, upc.as_upc_a()))

    return replace(result, upc=upc, upc_error=upc_error)


def _parse_sscc(
    value: str,
    config: ParseConfig,
    queue: ParseQueue,  # noqa: ARG001
    result: ParseResult,
) -> ParseResult:
    if result.sscc is not None:
        return result  # pragma: no cover

    try:
        sscc = Sscc.parse(value, config=config)
        sscc_error = None
    except ParseError as exc:
        sscc = None
        sscc_error = str(exc)

    return replace(result, sscc=sscc, sscc_error=sscc_error)


def _parse_gs1_message(
    value: str,
    config: ParseConfig,
    queue: ParseQueue,
    result: ParseResult,
) -> ParseResult:
    if result.gs1_message is not None:
        return result  # pragma: no cover

    try:
        gs1_message = GS1Message.parse(value, config=config)
        gs1_message_error = None
    except ParseError as exc:
        gs1_message = None
        gs1_message_error = str(exc)

    if gs1_message is not None:
        _promote_gs1_elements(gs1_message.element_strings, queue)

    return replace(result, gs1_message=gs1_message, gs1_message_error=gs1_message_error)


def _parse_gs1_web_uri(
    value: str,
    config: ParseConfig,
    queue: ParseQueue,
    result: ParseResult,
) -> ParseResult:
    if result.gs1_web_uri is not None:
        return result  # pragma: no cover

    try:
        gs1_web_uri = GS1WebURI.parse(value, config=config)
        gs1_web_uri_error = None
    except ParseError as exc:
        gs1_web_uri = None
        gs1_web_uri_error = str(exc)

    if gs1_web_uri is not None:
        _promote_gs1_elements(gs1_web_uri.element_strings, queue)

    return replace(result, gs1_web_uri=gs1_web_uri, gs1_web_uri_error=gs1_web_uri_error)


def _promote_gs1_elements(
    gs1_element_strings: GS1ElementStrings,
    queue: ParseQueue,
) -> None:
    # If the GS1 Message contains an SSCC, set SSCC on the top-level result.
    ai_00 = gs1_element_strings.get(ai="00")
    if ai_00 is not None:
        queue.append((_parse_sscc, ai_00.value))

    # If the GS1 Message contains an GTIN, set GTIN on the top-level result.
    ai_01 = gs1_element_strings.get(ai="01")
    if ai_01 is not None:
        queue.append((_parse_gtin, ai_01.value))



"""Library's own exceptions."""

__all__ = ["BiipException", "EncodeError", "ParseError"]


class BiipException(Exception):  # noqa: N818
    """Base class for all custom exceptions raised by the library."""


class EncodeError(BiipException):
    """Error raised if encoding of a value fails."""


class ParseError(BiipException):
    """Error raised if parsing of barcode data fails."""



"""Tests of encoding UPCs."""

import pytest

from biip import EncodeError
from biip.upc import Upc


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (  # UPC-A to UPC-A
            "123456789012",
            "123456789012",
        ),
        (  # Reverse of UPC-A suppression, condition A
            "02345673",
            "023456000073",
        ),
        (  # Reverse of UPC-A suppression, condition B
            "02345147",
            "023450000017",
        ),
        (  # Reverse of UPC-A suppression, condition C
            "06397126",
            "063200009716",
        ),
        (  # Reverse of UPC-A suppression, condition D
            "08679339",
            "086700000939",
        ),
    ],
)
def test_as_upc_a(value: str, expected: str) -> None:
    assert Upc.parse(value).as_upc_a() == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (  # 6-digit UPC-E, implicit number system 0, no check digit.
            "425261",
            "04252614",
        ),
        (  # 7-digit UPC-E, explicit number system 1, no check digit.
            "1425261",
            "14252611",
        ),
        (  # 8-digit UPC-E, explicit number system 1, with check digit.
            "14252611",
            "14252611",
        ),
        (  # UPC-A suppression, condition A
            "023456000073",
            "02345673",
        ),
        (  # UPC-A suppression, condition B
            "023450000017",
            "02345147",
        ),
        (  # UPC-A suppression, condition C
            "063200009716",
            "06397126",
        ),
        (  # UPC-A suppression, condition D
            "086700000939",
            "08679339",
        ),
    ],
)
def test_as_upc_e(value: str, expected: str) -> None:
    assert Upc.parse(value).as_upc_e() == expected


def test_as_upc_e_when_suppression_is_not_possible() -> None:
    with pytest.raises(EncodeError) as exc_info:
        Upc.parse("123456789012").as_upc_e()

    assert str(exc_info.value) == "UPC-A '123456789012' cannot be represented as UPC-E."


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("425261", "042100005264"),  # UPC-E, 6 digit
        ("0425261", "042100005264"),  # UPC-E, 7 digit
        ("04252614", "042100005264"),  # UPC-E, 8 digit
        ("042100005264", "042100005264"),  # UPC-A
    ],
)
def test_as_gtin_12(value: str, expected: str) -> None:
    assert Upc.parse(value).as_gtin_12() == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("425261", "0042100005264"),  # UPC-E, 6 digit
        ("0425261", "0042100005264"),  # UPC-E, 7 digit
        ("04252614", "0042100005264"),  # UPC-E, 8 digit
        ("042100005264", "0042100005264"),  # UPC-A
    ],
)
def test_as_gtin_13(value: str, expected: str) -> None:
    assert Upc.parse(value).as_gtin_13() == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("425261", "00042100005264"),  # UPC-E, 6 digit
        ("0425261", "00042100005264"),  # UPC-E, 7 digit
        ("04252614", "00042100005264"),  # UPC-E, 8 digit
        ("042100005264", "00042100005264"),  # UPC-A
    ],
)
def test_as_gtin_14(value: str, expected: str) -> None:
    assert Upc.parse(value).as_gtin_14() == expected



"""Checksum algorithms used by GS1 standards."""

import itertools


def gs1_standard_check_digit(value: str) -> int:
    """Get GS1 check digit for numeric string.

    Args:
        value: The numeric string to calculate the check digit for.

    Returns:
        The check digit.

    Raises:
        ValueError: If the value isn't numeric.

    References:
        GS1 General Specification, section 7.9.1

    Examples:
        >>> from biip.checksums import gs1_standard_check_digit
        >>> gs1_standard_check_digit("950110153100")  # GTIN-13
        0
        >>> gs1_standard_check_digit("9501234")  # GTIN-8
        6
    """
    if not value.isdecimal():
        msg = f"Expected numeric value, got {value!r}."
        raise ValueError(msg)

    digits = list(map(int, list(value)))
    reversed_digits = reversed(digits)

    weighted_sum = 0
    for digit, weight in zip(reversed_digits, itertools.cycle([3, 1])):
        weighted_sum += digit * weight

    return (10 - weighted_sum % 10) % 10


def gs1_price_weight_check_digit(value: str) -> int:
    """Get GS1 check digit for a price or weight field.

    Args:
        value: The numeric string to calculate the check digit for.

    Returns:
        The check digit.

    Raises:
        ValueError: If the value isn't numeric.

    References:
        GS1 General Specification, section 7.9.2-7.9.4

    Examples:
        >>> from biip.checksums import gs1_price_weight_check_digit
        >>> gs1_price_weight_check_digit("2875")
        9
        >>> gs1_price_weight_check_digit("14685")
        6
    """
    if not value.isdecimal():
        msg = f"Expected numeric value, got {value!r}."
        raise ValueError(msg)

    if len(value) == 4:
        return _four_digit_price_weight_check_digit(value)

    if len(value) == 5:
        return _five_digit_price_weight_check_digit(value)

    msg = f"Expected input of length 4 or 5, got {value!r}."
    raise ValueError(msg)


def _four_digit_price_weight_check_digit(value: str) -> int:
    digits = list(map(int, list(value)))
    weight_sum = 0
    for digit, weight_map in zip(digits, _FOUR_DIGIT_POSITION_WEIGHTS, strict=True):
        weight = weight_map[digit]
        weight_sum += weight
    return (weight_sum * 3) % 10


def _five_digit_price_weight_check_digit(value: str) -> int:
    digits = list(map(int, list(value)))
    weighted_sum = 0
    for digit, weight_map in zip(digits, _FIVE_DIGIT_POSITION_WEIGHTS, strict=True):
        weight = weight_map[digit]
        weighted_sum += weight
    result = (10 - weighted_sum % 10) % 10
    return _FIVE_MINUS_WEIGHT_REVERSE[result]


# See GS1 General Specification, section 7.9.2 for details.
_TWO_MINUS_WEIGHT = {0: 0, 1: 2, 2: 4, 3: 6, 4: 8, 5: 9, 6: 1, 7: 3, 8: 5, 9: 7}
_THREE_WEIGHT = {0: 0, 1: 3, 2: 6, 3: 9, 4: 2, 5: 5, 6: 8, 7: 1, 8: 4, 9: 7}
_FIVE_PLUS_WEIGHT = {0: 0, 1: 5, 2: 1, 3: 6, 4: 2, 5: 7, 6: 3, 7: 8, 8: 4, 9: 9}
_FIVE_MINUS_WEIGHT = {
    0: 0,
    1: 5,
    2: 9,
    3: 4,
    4: 8,
    5: 3,
    6: 7,
    7: 2,
    8: 6,
    9: 1,
}
_FIVE_MINUS_WEIGHT_REVERSE = {v: k for k, v in _FIVE_MINUS_WEIGHT.items()}

# See GS1 General Specification, section 7.9.3 for details.
_FOUR_DIGIT_POSITION_WEIGHTS = [
    _TWO_MINUS_WEIGHT,
    _TWO_MINUS_WEIGHT,
    _THREE_WEIGHT,
    _FIVE_MINUS_WEIGHT,
]

# See GS1 General Specification, section 7.9.4 for details.
_FIVE_DIGIT_POSITION_WEIGHTS = [
    _FIVE_PLUS_WEIGHT,
    _TWO_MINUS_WEIGHT,
    _FIVE_MINUS_WEIGHT,
    _FIVE_PLUS_WEIGHT,
    _TWO_MINUS_WEIGHT,
]



import pytest

from biip import EncodeError, ParseConfig
from biip.checksums import gs1_standard_check_digit
from biip.gtin import Gtin
from biip.rcn import Rcn, RcnRegion


@pytest.mark.parametrize(
    ("rcn_region", "value", "expected"),
    [
        (RcnRegion.DENMARK, "2712341002251", "2712341000004"),  # GTIN-13
        (RcnRegion.DENMARK, "02712341002251", "2712341000004"),  # GTIN-14
        (RcnRegion.ESTONIA, "2311111112345", "2311111100007"),
        (RcnRegion.FINLAND, "2311111112345", "2311111100007"),
        (RcnRegion.GERMANY, "2811111068708", "2811110000006"),  # GTIN-13
        (RcnRegion.GERMANY, "02811111068708", "2811110000006"),  # GTIN-14
        (RcnRegion.GREAT_BRITAIN, "2011122912346", "2011122000005"),
        (RcnRegion.LATVIA, "2311111112345", "2311111100007"),
        (RcnRegion.LITHUANIA, "2311111112345", "2311111100007"),
        (RcnRegion.NORWAY, "2302148210869", "2302148200006"),
        (RcnRegion.SWEDEN, "2088060112343", "2088060100005"),
    ],
)
def test_without_variable_measure_strips_variable_parts(
    rcn_region: RcnRegion, value: str, expected: str
) -> None:
    original_rcn = Gtin.parse(value, config=ParseConfig(rcn_region=rcn_region))
    assert isinstance(original_rcn, Rcn)

    stripped_rcn = original_rcn.without_variable_measure()

    assert isinstance(stripped_rcn, Rcn)
    assert stripped_rcn.as_gtin_13() == expected
    assert stripped_rcn.region == original_rcn.region


@pytest.mark.parametrize(
    ("rcn_region", "nonvariable_prefixes"),
    [
        (
            RcnRegion.DENMARK,
            ["02", "20", "25", "29"],
        ),
        (
            RcnRegion.ESTONIA,
            ["02", "20", "21", "22", "26", "27", "28", "29"],
        ),
        (
            RcnRegion.FINLAND,
            ["02", "20", "21", "22", "26", "27", "28", "29"],
        ),
        (
            RcnRegion.GERMANY,
            ["02", "20", "21", "24", "27"],
        ),
        (
            RcnRegion.GREAT_BRITAIN,
            ["21", "22", "23", "24", "25", "26", "27", "28", "29"],
        ),
        (
            RcnRegion.LATVIA,
            ["02", "20", "21", "22", "26", "27", "28", "29"],
        ),
        (
            RcnRegion.LITHUANIA,
            ["02", "20", "21", "22", "26", "27", "28", "29"],
        ),
        (
            RcnRegion.NORWAY,
            ["02", "26", "27", "28", "29"],
        ),
        (
            RcnRegion.SWEDEN,
            ["02", "26", "27", "28", "29"],
        ),
    ],
)
def test_without_variable_measure_keeps_nonvariable_rcn_unchanged(
    rcn_region: RcnRegion, nonvariable_prefixes: list[str]
) -> None:
    for prefix in nonvariable_prefixes:
        payload = f"{prefix}1111111111"
        value = f"{payload}{gs1_standard_check_digit(payload)}"
        original_rcn = Gtin.parse(value, config=ParseConfig(rcn_region=rcn_region))
        assert isinstance(original_rcn, Rcn)

        stripped_rcn = original_rcn.without_variable_measure()

        assert isinstance(stripped_rcn, Rcn)
        assert stripped_rcn.value == original_rcn.value
        assert stripped_rcn.region == original_rcn.region


@pytest.mark.parametrize(
    ("rcn_region", "value"),
    [
        (RcnRegion.NORWAY, "00012348"),  # GTIN-8
        (RcnRegion.NORWAY, "412345678903"),  # GTIN-12
        (RcnRegion.NORWAY, "0412345678903"),  # GTIN-13
        (RcnRegion.NORWAY, "00412345678903"),  # GTIN-14
    ],
)
def test_without_variable_measure_keeps_company_rcn_unchanged(
    rcn_region: RcnRegion, value: str
) -> None:
    original_rcn = Gtin.parse(value, config=ParseConfig(rcn_region=rcn_region))
    assert isinstance(original_rcn, Rcn)

    stripped_rcn = original_rcn.without_variable_measure()

    assert isinstance(stripped_rcn, Rcn)
    assert stripped_rcn.value == original_rcn.value
    assert stripped_rcn.region == original_rcn.region


@pytest.mark.parametrize(
    "value",
    [
        "96385074",  # GTIN-8
        "614141000036",  # GTIN-12
        "5901234123457",  # GTIN-13
        "98765432109213",  # GTIN-14
    ],
)
def test_without_variable_measure_keeps_gtin_unchanged(value: str) -> None:
    original_gtin = Gtin.parse(value)
    assert isinstance(original_gtin, Gtin)
    assert not isinstance(original_gtin, Rcn)

    stripped_gtin = original_gtin.without_variable_measure()

    assert isinstance(stripped_gtin, Gtin)
    assert not isinstance(stripped_gtin, Rcn)
    assert stripped_gtin.value == original_gtin.value


def test_without_variable_measure_fails_if_rules_are_unknown() -> None:
    rcn = Gtin.parse("2302148210869", config=ParseConfig(rcn_region=None))
    assert isinstance(rcn, Rcn)

    with pytest.raises(EncodeError) as exc_info:
        rcn.without_variable_measure()

    assert str(exc_info.value) == (
        "Cannot zero out the variable measure part of '2302148210869' "
        "as the RCN rules for the geographical region None are unknown."
    )



import pytest

from biip import ParseError
from biip.gs1_application_identifiers import GS1ApplicationIdentifier


@pytest.mark.parametrize("unknown_ai", ["abcdef", "3376999999"])
def test_extract_unknown_gs1_ai(unknown_ai: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        GS1ApplicationIdentifier.extract(unknown_ai)

    assert (
        str(exc_info.value)
        == f"Failed to get GS1 Application Identifier from {unknown_ai!r}."
    )


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "0195012345678903",
            GS1ApplicationIdentifier(
                ai="01",
                description="Global Trade Item Number (GTIN)",
                data_title="GTIN",
                separator_required=False,
                format="N2+N14",
                pattern=r"^01(\d{14})$",
            ),
        ),
        (
            "90SE011\x1d2501611262",
            GS1ApplicationIdentifier(
                ai="90",
                description="Information mutually agreed between trading partners",
                data_title="INTERNAL",
                separator_required=True,
                format="N2+X..30",
                pattern=r"^90([!%-?A-Z_a-z\x22]{1,30})$",
            ),
        ),
    ],
)
def test_extract_gs1_ai(value: str, expected: GS1ApplicationIdentifier) -> None:
    assert GS1ApplicationIdentifier.extract(value) == expected


def test_gs1_ai_object_len_is_ai_str_len() -> None:
    ai = GS1ApplicationIdentifier.extract("01")

    assert len(ai) == len("01")


def test_is_hashable() -> None:
    ai = GS1ApplicationIdentifier.extract("01")

    assert hash(ai) is not None



"""Restricted Circulation Number (RCN) is a subset of GTIN.

Both RCN-8, RCN-12, and RCN-13 are supported. There is no 14 digit version
of RCN.

RCN-12 with prefix 2 and RCN-13 with prefix 02 or 20-29 have the same
semantics across a geographic region, defined by the local GS1 Member
Organization.

RCN-8 with prefix 0 or 2, RCN-12 with prefix 4, and RCN-13 with prefix 04 or
40-49 have semantics that are only defined within a single company.

Use [`Gtin.parse()`][biip.gtin.Gtin.parse] to parse potential RCNs. This
subclass is returned if the GS1 Prefix signifies that the value is an RCN.

References:
    GS1 General Specifications, section 2.1.11-2.1.12
"""

from __future__ import annotations

from dataclasses import dataclass, field, replace
from decimal import Decimal
from enum import Enum
from typing import TYPE_CHECKING, Any

from biip import EncodeError, ParseError
from biip._parser import ParseConfig
from biip.checksums import gs1_price_weight_check_digit, gs1_standard_check_digit
from biip.gtin import Gtin

if TYPE_CHECKING:
    from collections.abc import Iterator

try:
    import moneyed  # noqa: TC002

    have_moneyed = True
except ImportError:  # pragma: no cover
    have_moneyed = False

__all__ = ["Rcn", "RcnRegion", "RcnUsage"]


class RcnUsage(Enum):
    """Enum of RCN usage restrictions."""

    GEOGRAPHICAL = "geo"
    """Usage of RCN restricted to geopgraphical area."""

    COMPANY = "company"
    """Usage of RCN restricted to internally in a company."""

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"RcnUsage.{self.name}"


class RcnRegion(Enum):
    """Enum of geographical regions with custom RCN rules.

    The value of the enum is the lowercase ISO 3166-1 Alpha-2 code.
    """

    DENMARK = "dk"
    """Denmark"""

    ESTONIA = "ee"
    """Estonia"""

    FINLAND = "fi"
    """Finland"""

    GERMANY = "de"
    """Germany"""

    GREAT_BRITAIN = "gb"
    """Great Britain"""

    LATVIA = "lv"
    """Latvia"""

    LITHUANIA = "lt"
    """Lithuania"""

    NORWAY = "no"
    """Norway"""

    SWEDEN = "se"
    """Sweden"""

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"RcnRegion.{self.name}"

    def get_currency_code(self) -> str | None:
        """Get the ISO-4217 currency code for the region."""
        return {
            RcnRegion.DENMARK: "DKK",
            RcnRegion.GERMANY: "EUR",
            RcnRegion.GREAT_BRITAIN: "GBP",
            RcnRegion.NORWAY: "NOK",
            RcnRegion.SWEDEN: "SEK",
        }.get(self)


@dataclass(frozen=True)
class Rcn(Gtin):
    """Data class containing an RCN.

    This is a subclass of [`Gtin`][biip.gtin.Gtin]. To create an `Rcn` instance,
    use [`Gtin.parse()`][biip.gtin.Gtin.parse] with an RCN string as the value
    to parse.
    """

    usage: RcnUsage | None = field(default=None)
    """Where the RCN can be circulated, in a geographical region or within a company."""

    region: RcnRegion | None = field(default=None)
    """The geographical region.

    The region's rules are used to interpret the contents of the RCN.
    """

    weight: Decimal | None = field(default=None)
    """A variable weight value extracted from the GTIN."""

    count: int | None = field(default=None)
    """A variable count extracted from the GTIN."""

    price: Decimal | None = field(default=None)
    """A variable weight price extracted from the GTIN."""

    money: moneyed.Money | None = field(default=None)
    """A Money value created from the variable weight price.

    Only set if [`py-moneyed`](https://pypi.org/project/py-moneyed/) is
    installed and the currency is known.
    """

    def __rich_repr__(self) -> Iterator[tuple[str, Any] | tuple[str, Any, Any]]:  # noqa: D105
        # Skip printing fields with default values
        yield from super().__rich_repr__()
        yield "usage", self.usage, None
        yield "region", self.region, None
        yield "weight", self.weight, None
        yield "count", self.count, None
        yield "price", self.price, None
        yield "money", self.money, None

    def _with_usage(self) -> Rcn:
        # Classification as RCN depends on the prefix being known, so we won't
        # get here unless it is known.
        assert self.prefix is not None

        usage = None
        if "within a geographic region" in self.prefix.usage:
            usage = RcnUsage.GEOGRAPHICAL
        if "within a company" in self.prefix.usage:
            usage = RcnUsage.COMPANY

        return replace(self, usage=usage)

    def _parsed_with_regional_rules(self, *, config: ParseConfig) -> Rcn:
        rcn = self

        if rcn.usage == RcnUsage.COMPANY:
            # The value is an RCN, but it is intended for use within a company,
            # so we can only interpret it as an opaque GTIN.
            return rcn

        if config.rcn_region is None:
            # The region is not known, so we cannot interpret the variable measure.
            return rcn

        rcn = replace(rcn, region=RcnRegion(config.rcn_region))

        strategy = _Strategy.get_for_rcn(rcn)
        if strategy is None:
            # Without a strategy, we cannot extract anything.
            return rcn

        if config.rcn_verify_variable_measure:
            strategy.verify_check_digit(rcn)

        if strategy.measure_type == _MeasureType.WEIGHT:
            weight = strategy.get_variable_measure(self)
        else:
            weight = None

        if strategy.measure_type == _MeasureType.COUNT:
            count = int(strategy.get_variable_measure(self))
        else:
            count = None

        if strategy.measure_type == _MeasureType.PRICE:
            price = strategy.get_variable_measure(self)
        else:
            price = None

        assert rcn.region
        currency_code = rcn.region.get_currency_code()
        if price is not None and have_moneyed and currency_code is not None:
            import moneyed

            money = moneyed.Money(amount=price, currency=currency_code)
        else:
            money = None

        return replace(
            rcn,
            weight=weight,
            count=count,
            price=price,
            money=money,
        )

    def without_variable_measure(self) -> Gtin:
        """Create a new RCN where the variable measure is zeroed out.

        This provides us with a number which still includes the item
        reference, but does not vary with weight/price, and can thus be used
        to lookup the relevant trade item in a database or similar.

        This has no effect on RCNs intended for use within a company, as
        the semantics of those numbers vary from company to company.

        Returns:
            A RCN instance with zeros in the variable measure places.

        Raises:
            EncodeError: If the rules for variable measures in the region are unknown.
        """
        if self.usage == RcnUsage.COMPANY:
            # The value is an RCN, but it is intended for use within a company,
            # so we can only interpret it as an opaque GTIN.
            return self

        if self.region is None:
            msg = (
                f"Cannot zero out the variable measure part of {self.value!r} as the "
                f"RCN rules for the geographical region {self.region!r} are unknown."
            )
            raise EncodeError(msg)

        strategy = _Strategy.get_for_rcn(self)
        if strategy is None:
            # This prefix has no rules for removing variable parts.
            return self

        return strategy.without_variable_measure(self)


class _MeasureType(str, Enum):
    COUNT = "count"
    PRICE = "price"
    WEIGHT = "weight"


@dataclass
class _Strategy:
    measure_type: _MeasureType
    pattern: str
    num_decimals: int

    prefix_slice: slice = field(init=False)
    value_slice: slice = field(init=False)
    check_digit_slice: slice | None = field(init=False)

    @classmethod
    def get_for_rcn(cls, rcn: Rcn) -> _Strategy | None:
        # The RCN's geographical region must be known to lookup the correct
        # strategy for interpreting the RCN's variable measure.
        assert rcn.region is not None

        region_rules = _RCN_RULES.get(rcn.region)
        if region_rules is None:  # pragma: no cover
            msg = "RCN region defined without defining rules. This is a bug."
            raise Exception(msg)  # noqa: TRY002

        # Classification as RCN depends on the prefix being known, so we won't
        # get here unless it is known.
        assert rcn.prefix is not None

        rcn_prefix = rcn.prefix.value[:2]
        return region_rules.get(rcn_prefix)

    def __post_init__(self) -> None:
        assert len(self.pattern) == 12, "Pattern must be exactly 12 chars long."

        prefix_slice = self._get_pattern_slice("P")
        value_slice = self._get_pattern_slice("V")

        assert prefix_slice is not None, "Pattern must contain a prefix marker (P)."
        assert value_slice is not None, "Pattern must contain a value marker (V)."

        self.prefix_slice = prefix_slice
        self.value_slice = value_slice
        self.check_digit_slice = self._get_pattern_slice("C")

    def _get_pattern_slice(self, char: str) -> slice | None:
        if char not in self.pattern:
            return None
        return slice(self.pattern.index(char), self.pattern.rindex(char) + 1)

    def verify_check_digit(self, rcn: Rcn) -> None:
        if self.check_digit_slice is None:
            return

        rcn_13 = rcn.as_gtin_13()
        value = rcn_13[self.value_slice]
        check_digit = int(rcn_13[self.check_digit_slice])
        calculated_check_digit = gs1_price_weight_check_digit(value)

        if check_digit != calculated_check_digit:
            msg = (
                f"Invalid check digit for variable measure value {value!r} "
                f"in RCN {rcn.value!r}: "
                f"Expected {calculated_check_digit!r}, got {check_digit!r}."
            )
            raise ParseError(msg)

    def get_variable_measure(self, rcn: Rcn) -> Decimal:
        rcn_13 = rcn.as_gtin_13()
        value = Decimal(rcn_13[self.value_slice])
        return value / Decimal(10) ** self.num_decimals

    def without_variable_measure(self, rcn: Rcn) -> Gtin:
        # Zero out the variable measure part of the payload, and recalculate both
        # the GTIN check digit and the variable measure's check digit digit, if any.

        rcn_13 = rcn.as_gtin_13()
        zeroed_value = "0" * len(rcn_13[self.value_slice])

        digits = list(self.pattern)
        digits[self.prefix_slice] = list(rcn_13[self.prefix_slice])
        digits[self.value_slice] = list(zeroed_value)
        if self.check_digit_slice is not None:
            digits[self.check_digit_slice] = [
                str(gs1_price_weight_check_digit(zeroed_value))
            ]

        gtin_payload = "".join(digits)
        gtin_check_digit = gs1_standard_check_digit(gtin_payload)
        gtin = f"{gtin_payload}{gtin_check_digit}"
        return Gtin.parse(gtin, config=ParseConfig(rcn_region=rcn.region))


_RCN_RULES: dict[RcnRegion, dict[str, _Strategy]] = {
    RcnRegion.DENMARK: {
        # References:
        #   https://www.gs1.dk/om-gs1/overblik-over-gs1-standarder/gtin-13-pris
        #   https://www.gs1.dk/om-gs1/overblik-over-gs1-standarder/gtin-13-vaegt
        "21": _Strategy(_MeasureType.PRICE, "PPPPPPVVVVVV", num_decimals=2),
        "22": _Strategy(_MeasureType.PRICE, "PPPPPPVVVVVV", num_decimals=2),
        "23": _Strategy(_MeasureType.PRICE, "PPPPPPVVVVVV", num_decimals=2),
        "24": _Strategy(_MeasureType.PRICE, "PPPPPPVVVVVV", num_decimals=2),
        "26": _Strategy(_MeasureType.WEIGHT, "PPPPPPVVVVVV", num_decimals=3),
        "27": _Strategy(_MeasureType.WEIGHT, "PPPPPPPVVVVV", num_decimals=3),
        "28": _Strategy(_MeasureType.WEIGHT, "PPPPPPVVVVVV", num_decimals=3),
    },
    RcnRegion.ESTONIA: {
        # References:
        #   https://gs1lv.org/img/upload/ENG.Variable%20measure_in_Latvia.pdf
        "23": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=3),
        "24": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=2),
        "25": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=1),
    },
    RcnRegion.FINLAND: {
        # References:
        #   https://gs1.fi/en/instructions/gs1-company-prefix/how-identify-product-gtin
        "23": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=3),
        "24": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=2),
        "25": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=1),
    },
    RcnRegion.GERMANY: {
        # References:
        #   https://www.gs1-germany.de/fileadmin/gs1/fachpublikationen/globale-artikelnummer-gtin-in-der-anwendung.pdf
        #   https://san.gs1-germany.de/SAN-4-Konzept
        "22": _Strategy(_MeasureType.PRICE, "PPPPPPCVVVVV", num_decimals=2),
        "23": _Strategy(_MeasureType.PRICE, "PPPPPPCVVVVV", num_decimals=2),
        "25": _Strategy(_MeasureType.COUNT, "PPPPPPCVVVVV", num_decimals=0),
        "26": _Strategy(_MeasureType.COUNT, "PPPPPPCVVVVV", num_decimals=0),
        "28": _Strategy(_MeasureType.WEIGHT, "PPPPPPCVVVVV", num_decimals=3),
        "29": _Strategy(_MeasureType.WEIGHT, "PPPPPPCVVVVV", num_decimals=3),
    },
    RcnRegion.GREAT_BRITAIN: {
        # References:
        #   https://www.gs1uk.org/how-to-barcode-variable-measure-items
        "20": _Strategy(_MeasureType.PRICE, "PPPPPPPCVVVV", num_decimals=2),
    },
    RcnRegion.LATVIA: {
        # References:
        #   https://gs1lv.org/img/upload/ENG.Variable%20measure_in_Latvia.pdf
        "23": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=3),
        "24": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=2),
        "25": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=1),
    },
    RcnRegion.LITHUANIA: {
        # References:
        #   https://gs1lv.org/img/upload/ENG.Variable%20measure_in_Latvia.pdf
        "23": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=3),
        "24": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=2),
        "25": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=1),
    },
    RcnRegion.NORWAY: {
        # References:
        #   No specification found, but products tested seems to match Swedish rules.
        "20": _Strategy(_MeasureType.PRICE, "PPPPPPPPVVVV", num_decimals=2),
        "21": _Strategy(_MeasureType.PRICE, "PPPPPPPPVVVV", num_decimals=1),
        "22": _Strategy(_MeasureType.PRICE, "PPPPPPPPVVVV", num_decimals=0),
        "23": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=3),
        "24": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=2),
        "25": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=1),
    },
    RcnRegion.SWEDEN: {
        # References:
        #   https://gs1.se/en/support/how-do-i-create-my-variable-weight-numbers/
        "20": _Strategy(_MeasureType.PRICE, "PPPPPPPPVVVV", num_decimals=2),
        "21": _Strategy(_MeasureType.PRICE, "PPPPPPPPVVVV", num_decimals=1),
        "22": _Strategy(_MeasureType.PRICE, "PPPPPPPPVVVV", num_decimals=0),
        "23": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=3),
        "24": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=2),
        "25": _Strategy(_MeasureType.WEIGHT, "PPPPPPPPVVVV", num_decimals=1),
    },
}



"""Tests of parsing UPCs."""

import pytest

from biip import ParseError
from biip.upc import Upc, UpcFormat


def test_parse_upc_a() -> None:
    upc = Upc.parse("042100005264")

    assert upc == Upc(
        value="042100005264",
        format=UpcFormat.UPC_A,
        number_system_digit=0,
        payload="04210000526",
        check_digit=4,
    )


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "425261",  # Length is 6: Implicit number system 0, no check digit.
            Upc(
                value="425261",
                format=UpcFormat.UPC_E,
                number_system_digit=0,
                payload="0425261",
                check_digit=4,
            ),
        ),
        (
            "1425261",  # Length is 7: Explicit number system 1, no check digit.
            Upc(
                value="1425261",
                format=UpcFormat.UPC_E,
                number_system_digit=1,
                payload="1425261",
                check_digit=1,
            ),
        ),
        (
            "14252611",  # Length is 8: Explicit number system 1 and check digit.
            Upc(
                value="14252611",
                format=UpcFormat.UPC_E,
                number_system_digit=1,
                payload="1425261",
                check_digit=1,
            ),
        ),
    ],
)
def test_parse_upc_e(value: str, expected: Upc) -> None:
    assert Upc.parse(value) == expected


def test_parse_value_with_invalid_length() -> None:
    with pytest.raises(ParseError) as exc_info:
        Upc.parse("123")

    assert (
        str(exc_info.value)
        == "Failed to parse '123' as UPC: Expected 6, 7, 8, or 12 digits, got 3."
    )


@pytest.mark.parametrize("value", ["012345678abc", "012345678⁰⁰⁰"])
def test_parse_nonnumeric_value(value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        Upc.parse(value)

    assert (
        str(exc_info.value)
        == f"Failed to parse {value!r} as UPC: Expected a numerical value."
    )


def test_parse_upc_a_with_invalid_check_digit() -> None:
    with pytest.raises(ParseError) as exc_info:
        Upc.parse("042100005263")

    assert (
        str(exc_info.value)
        == "Invalid UPC-A check digit for '042100005263': Expected 4, got 3."
    )


def test_parse_upc_e_with_invalid_check_digit() -> None:
    with pytest.raises(ParseError) as exc_info:
        Upc.parse("04252613")

    assert (
        str(exc_info.value)
        == "Invalid UPC-E check digit for '04252613': Expected 4, got 3."
    )


@pytest.mark.parametrize("number_system", range(2, 10))
def test_parse_upc_e_with_invalid_number_system(number_system: int) -> None:
    with pytest.raises(ParseError) as exc_info:
        Upc.parse(f"{number_system}425261")

    assert str(exc_info.value) == (
        f"Invalid UPC-E number system for '{number_system}425261': "
        f"Expected 0 or 1, got {number_system}."
    )


def test_parse_strips_surrounding_whitespace() -> None:
    upc = Upc.parse("  \t 042100005264 \n  ")

    assert upc.value == "042100005264"



r"""Biip interprets the data in barcodes.

    >>> import biip

An ambiguous value may be interpreted as different formats. In the following
example, the value can be interpreted as either a GTIN or a GS1 Message.

    >>> result = biip.parse("96385074")
    >>> pprint(result.gtin)
    Gtin(
        value='96385074',
        format=GtinFormat.GTIN_8,
        prefix=GS1Prefix(
            value='00009',
            usage='GS1 US'
        ),
        company_prefix=GS1CompanyPrefix(
            value='0000963'
        ),
        payload='9638507',
        check_digit=4
    )
    >>> pprint(result.gs1_message)
    GS1Message(
        value='96385074',
        element_strings=[
            GS1ElementString(
                ai=GS1ApplicationIdentifier(
                    ai='96',
                    description='Company internal information',
                    data_title='INTERNAL',
                    separator_required=True,
                    format='N2+X..90'
                ),
                value='385074',
                pattern_groups=[
                    '385074'
                ]
            )
        ]
    )

In the next example, the value is only valid as a GS1 Message and the GTIN
parser returns an error explaining why the value cannot be interpreted as a
GTIN. If a format includes check digits, Biip always control them and fail if
the check digits are incorrect.

    >>> result = biip.parse("15210527")
    >>> pprint(result.gtin)
    None
    >>> result.gtin_error
    "Invalid GTIN check digit for '15210527': Expected 4, got 7."
    >>> pprint(result.gs1_message)
    GS1Message(
        value='15210527',
        element_strings=[
            GS1ElementString(
                ai=GS1ApplicationIdentifier(
                    ai='15',
                    description='Best before date (YYMMDD)',
                    data_title='BEST BEFORE or BEST BY',
                    separator_required=False,
                    format='N2+N6'
                ),
                value='210527',
                pattern_groups=[
                    '210527'
                ],
                date=datetime.date(2021, 5, 27)
            )
        ]
    )

If a value cannot be interpreted as any supported format, you still get a
[`ParseResult`][biip.ParseResult] object with all result fields set to `None`.

    >>> result = biip.parse("123")
    >>> pprint(result)
    ParseResult(
        value='123',
        gtin_error="Failed to parse '123' as GTIN: Expected 8, 12, 13, or 14 digits, got 3.",
        upc_error="Failed to parse '123' as UPC: Expected 6, 7, 8, or 12 digits, got 3.",
        sscc_error="Failed to parse '123' as SSCC: Expected 18 digits, got 3.",
        gs1_message_error="Failed to match '123' with GS1 AI (12) pattern '^12(\\d{2}(?:0\\d|1[0-2])(?:[0-2]\\d|3[01]))$'."
    )
"""  # noqa: E501

from importlib.metadata import (  # pyright: ignore[reportMissingImports]
    PackageNotFoundError,  # pyright: ignore[reportUnknownVariableType]
    version,  # pyright: ignore[reportUnknownVariableType]
)

from biip._config import ParseConfig
from biip._exceptions import BiipException, EncodeError, ParseError
from biip._parser import ParseResult, parse

__all__ = [
    "BiipException",
    "EncodeError",
    "ParseConfig",
    "ParseError",
    "ParseResult",
    "parse",
]


try:
    __version__: str = version(__name__)  # pyright: ignore[reportUnknownVariableType]
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable

    from biip.rcn import RcnRegion


@dataclass(frozen=True)
class ParseConfig:
    """Configuration options for parsers.

    Biip strives to do the right thing by default, but you can customize
    some of its the behavior by setting these options.
    """

    gs1_element_strings_verify_date: bool = True
    """Whether to verify that the date in a GS1 element string is valid.

    According to the GS1 General Specification, dates are required to contain a
    valid year and month. Only the day of month can be left as zeros, which
    should be interpreted as the last day of the month.

    Some barcodes include a GS1 element string with all zero dates, requiring
    this check to be disabled.
    """

    rcn_region: RcnRegion | str | None = None
    """The geographical region to use when parsing RCNs.

    Restricted Circulation Numbers (RCN) have different meanings in different
    geographical regions. Specify your region here so that Biip can use the
    right parsing rules.

    This must be set to extract variable weight or price from GTINs.
    """

    rcn_verify_variable_measure: bool = True
    """
    Whether to verify that the variable measure in a RCN matches its check digit.

    Some companies use the variable measure check digit for other purposes,
    requiring this check to be disabled.
    """

    separator_chars: Iterable[str] = ("\x1d",)
    """
    Characters to look for as separators between fields in GS1 messages.

    Defaults to the "group separator" character (byte `0x1d` in ASCII), which is
    commonly returned by barcode scanners in place of the FNC1 barcode symbol
    that separates fields.

    If variable-length fields in the middle of the message are not terminated
    with a separator character, the parser might greedily consume the rest of
    the message.
    """



import datetime as dt
from decimal import Decimal

import pytest

from biip import ParseError
from biip.gln import Gln
from biip.gs1_application_identifiers import GS1ApplicationIdentifier
from biip.gs1_element_strings import GS1ElementString, GS1ElementStrings
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix
from biip.gtin import Gtin, GtinFormat
from biip.sscc import Sscc


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "00373400306809981733",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("00"),
                value="373400306809981733",
                pattern_groups=["373400306809981733"],
                sscc=Sscc(
                    value="373400306809981733",
                    prefix=GS1Prefix(value="734", usage="GS1 Sweden"),
                    company_prefix=GS1CompanyPrefix(value="73400306"),
                    extension_digit=3,
                    payload="37340030680998173",
                    check_digit=3,
                ),
            ),
        ),
        (
            "0107032069804988",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("01"),
                value="07032069804988",
                pattern_groups=["07032069804988"],
                gtin=Gtin(
                    value="07032069804988",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="703", usage="GS1 Norway"),
                    company_prefix=GS1CompanyPrefix(value="703206"),
                    payload="703206980498",
                    check_digit=8,
                ),
            ),
        ),
        (
            "0207032069804988",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("02"),
                value="07032069804988",
                pattern_groups=["07032069804988"],
                gtin=Gtin(
                    value="07032069804988",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="703", usage="GS1 Norway"),
                    company_prefix=GS1CompanyPrefix(value="703206"),
                    payload="703206980498",
                    check_digit=8,
                ),
            ),
        ),
        (
            "0307032069804988",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("03"),
                value="07032069804988",
                pattern_groups=["07032069804988"],
                gtin=Gtin(
                    value="07032069804988",
                    format=GtinFormat.GTIN_13,
                    prefix=GS1Prefix(value="703", usage="GS1 Norway"),
                    company_prefix=GS1CompanyPrefix(value="703206"),
                    payload="703206980498",
                    check_digit=8,
                ),
            ),
        ),
        (
            "100329",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("10"),
                value="0329",
                pattern_groups=["0329"],
            ),
        ),
        (
            "4101234567890128",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("410"),
                value="1234567890128",
                pattern_groups=["1234567890128"],
                gln=Gln(
                    value="1234567890128",
                    prefix=GS1Prefix(value="123", usage="GS1 US"),
                    company_prefix=GS1CompanyPrefix(value="1234567890"),
                    payload="123456789012",
                    check_digit=8,
                ),
            ),
        ),
        (
            "7011030102",  # Without optional pattern group for time
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("7011"),
                value="030102",
                pattern_groups=["030102"],
                date=dt.date(2003, 1, 2),
                datetime=None,
            ),
        ),
        (
            "70110301021430",  # With optional pattern group for time
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("7011"),
                value="0301021430",
                pattern_groups=["030102", "1430"],
                date=dt.date(2003, 1, 2),
                datetime=dt.datetime(2003, 1, 2, 14, 30),  # noqa: DTZ001
            ),
        ),
        (
            "800307071324001022085952",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("8003"),
                value="07071324001022085952",
                pattern_groups=["0", "7071324001022", "085952"],
            ),
        ),
    ],
)
def test_extract(value: str, expected: GS1ElementString) -> None:
    assert GS1ElementString.extract(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (  # GS1 element string with invalid GLN
            "4101234567890127",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("410"),
                value="1234567890127",
                pattern_groups=["1234567890127"],
                gln=None,  # Not set, because GLN is invalid.
                gln_error=(
                    "Invalid GLN check digit for '1234567890127': Expected 8, got 7."
                ),
            ),
        ),
        (  # GS1 element string with invalid GTIN
            "0107032069804987",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("01"),
                value="07032069804987",
                pattern_groups=["07032069804987"],
                gtin=None,  # Not set, because GTIN is invalid.
                gtin_error=(
                    "Invalid GTIN check digit for '07032069804987': Expected 8, got 7."
                ),
            ),
        ),
        (  # GS1 element string with invalid SSCC
            "00376130321109103421",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("00"),
                value="376130321109103421",
                pattern_groups=["376130321109103421"],
                sscc=None,  # Not set, because SSCC is invalid.
                sscc_error=(
                    "Invalid SSCC check digit for '376130321109103421': "
                    "Expected 0, got 1."
                ),
            ),
        ),
    ],
)
def test_extract_with_nested_error(value: str, expected: GS1ElementString) -> None:
    assert GS1ElementString.extract(value) == expected


@pytest.mark.parametrize(
    ("ai_code", "bad_value"),
    [
        # Too short product number
        ("01", "01123"),
        # Too short weight
        ("3100", "3100123"),
    ],
)
def test_extract_fails_when_not_matching_pattern(ai_code: str, bad_value: str) -> None:
    ai = GS1ApplicationIdentifier.extract(ai_code)

    with pytest.raises(ParseError) as exc_info:
        GS1ElementString.extract(bad_value)

    assert (
        str(exc_info.value)
        == f"Failed to match {bad_value!r} with GS1 AI {ai} pattern '{ai.pattern}'."
    )


@pytest.mark.parametrize(
    ("value", "expected_date", "expected_datetime"),
    [
        ("11030201", dt.date(2003, 2, 1), None),
        ("12030201", dt.date(2003, 2, 1), None),
        ("13030201", dt.date(2003, 2, 1), None),
        ("15030201", dt.date(2003, 2, 1), None),
        ("16030201", dt.date(2003, 2, 1), None),
        ("17030201", dt.date(2003, 2, 1), None),
        ("43240701029999", dt.date(2007, 1, 2), None),
        ("43240701021430", dt.date(2007, 1, 2), dt.datetime(2007, 1, 2, 14, 30)),  # noqa: DTZ001
        ("43250701029999", dt.date(2007, 1, 2), None),
        ("43250701021430", dt.date(2007, 1, 2), dt.datetime(2007, 1, 2, 14, 30)),  # noqa: DTZ001
        ("4326070102", dt.date(2007, 1, 2), None),
        ("70030701021415", dt.date(2007, 1, 2), dt.datetime(2007, 1, 2, 14, 15)),  # noqa: DTZ001
        ("7006030102", dt.date(2003, 1, 2), None),
        ("7007030102", dt.date(2003, 1, 2), None),
        ("7011030102", dt.date(2003, 1, 2), None),
        ("70110301021430", dt.date(2003, 1, 2), dt.datetime(2003, 1, 2, 14, 30)),  # noqa: DTZ001
        ("800800010214", dt.date(2000, 1, 2), dt.datetime(2000, 1, 2, 14, 0)),  # noqa: DTZ001
        ("80080001021415", dt.date(2000, 1, 2), dt.datetime(2000, 1, 2, 14, 15)),  # noqa: DTZ001
        ("8008000102141516", dt.date(2000, 1, 2), dt.datetime(2000, 1, 2, 14, 15, 16)),  # noqa: DTZ001
    ],
)
def test_extract_date_and_datetime(
    value: str, expected_date: dt.date, expected_datetime: dt.datetime | None
) -> None:
    element_string = GS1ElementString.extract(value)

    assert element_string.date == expected_date
    assert element_string.datetime == expected_datetime


@pytest.mark.parametrize(
    ("ai_code", "bad_value"),
    [
        # Bad production date
        ("11", "120230"),
        # Bad best before date
        ("15", "990229"),
    ],
)
def test_extract_fails_with_invalid_date(ai_code: str, bad_value: str) -> None:
    ai = GS1ApplicationIdentifier.extract(ai_code)

    with pytest.raises(ParseError) as exc_info:
        GS1ElementString.extract(f"{ai_code}{bad_value}")

    assert (
        str(exc_info.value)
        == f"Failed to parse GS1 AI {ai} date/time from {bad_value!r}."
    )


THIS_YEAR = dt.datetime.now(tz=dt.timezone.utc).year
THIS_YEAR_SHORT = str(THIS_YEAR)[2:]
MIN_YEAR = THIS_YEAR - 49
MIN_YEAR_SHORT = str(MIN_YEAR)[2:]
MAX_YEAR = THIS_YEAR + 50
MAX_YEAR_SHORT = str(MAX_YEAR)[2:]


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            # Best before date, around the current date
            f"15{THIS_YEAR_SHORT}0526",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("15"),
                value=f"{THIS_YEAR_SHORT}0526",
                pattern_groups=[f"{THIS_YEAR_SHORT}0526"],
                date=dt.date(THIS_YEAR, 5, 26),
            ),
        ),
        (
            # Best before date, 49 years into the past
            f"15{MIN_YEAR_SHORT}0526",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("15"),
                value=f"{MIN_YEAR_SHORT}0526",
                pattern_groups=[f"{MIN_YEAR_SHORT}0526"],
                date=dt.date(MIN_YEAR, 5, 26),
            ),
        ),
        (
            # Best before date, 50 years into the future
            f"15{MAX_YEAR_SHORT}0526",
            GS1ElementString(
                ai=GS1ApplicationIdentifier.extract("15"),
                value=f"{MAX_YEAR_SHORT}0526",
                pattern_groups=[f"{MAX_YEAR_SHORT}0526"],
                date=dt.date(MAX_YEAR, 5, 26),
            ),
        ),
    ],
)
def test_extract_handles_min_and_max_year_correctly(
    value: str, expected: GS1ElementString
) -> None:
    assert GS1ElementString.extract(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("15200200", dt.date(2020, 2, 29)),
        ("15210200", dt.date(2021, 2, 28)),
        ("17211200", dt.date(2021, 12, 31)),
    ],
)
def test_extract_handles_zero_day_as_last_day_of_month(
    value: str, expected: dt.date
) -> None:
    assert GS1ElementString.extract(value).date == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        # Trade measures (GS1 General Specifications, section 3.6.2)
        ("3105123456", Decimal("1.23456")),  # Net weight (kg)
        ("3114123456", Decimal("12.3456")),  # First dimension (m)
        ("3123123456", Decimal("123.456")),  # Second dimension (m)
        ("3132123456", Decimal("1234.56")),  # Third dimension (m)
        ("3141123456", Decimal("12345.6")),  # Area (m^2)
        ("3150123456", Decimal("123456")),  # Net volume (l)
        ("3161123456", Decimal("12345.6")),  # Net volume (m^3)
        # ... plus equivalent for imperial units
        ("3661123456", Decimal("12345.6")),  # Net volume (cubic yards)
        #
        # Logistic measures (GS1 General Specifications, section 3.6.3)
        ("3302023456", Decimal("234.56")),  # Logistic weight (kg)
        ("3313023456", Decimal("23.456")),  # First dimension (m)
        ("3324023456", Decimal("2.3456")),  # Second dimension (m)
        ("3335023456", Decimal("0.23456")),  # Third dimension (m)
        ("3344023456", Decimal("2.3456")),  # Area (m^2)
        ("3353023456", Decimal("23.456")),  # Logistic volume (l)
        ("3362023456", Decimal("234.56")),  # Logistic volume (m^3)
        # ... plus equivalent for imperial units
        ("3691123456", Decimal("12345.6")),  # Logistic volume (cubic yards)
        #
        # Kilograms per square meter (GS1 General Specifications, section 3.6.4)
        ("3372123456", Decimal("1234.56")),
    ],
)
def test_extract_variable_measures(value: str, expected: Decimal) -> None:
    assert GS1ElementString.extract(value).decimal == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        # Amount payable or coupon value (GS1 General Specifications, section 3.6.6)
        ("3901123", Decimal("12.3")),
        ("3901123456", Decimal("12345.6")),
        ("3903123456789012345", Decimal("123456789012.345")),
        ("3909123456789012345", Decimal("123456.789012345")),
        # Amount payable for variable measure trade item (section 3.6.8)
        ("3921123", Decimal("12.3")),
        ("3921123456", Decimal("12345.6")),
        ("3923123456789012345", Decimal("123456789012.345")),
        ("3929123456789012345", Decimal("123456.789012345")),
    ],
)
def test_extract_amount_payable(value: str, expected: Decimal) -> None:
    assert GS1ElementString.extract(value).decimal == expected


@pytest.mark.parametrize(
    ("value", "expected_currency", "expected_decimal"),
    [
        # Amount payable and ISO currency code (section 3.6.7)
        ("39127101230", "ZAR", Decimal("12.30")),
        ("39117101230", "ZAR", Decimal("123.0")),
        ("391097812301", "EUR", Decimal("12301")),
        #
        # Amount payable for variable measure trade item and currency (section 3.6.9)
        ("39327101230", "ZAR", Decimal("12.30")),
        ("39317101230", "ZAR", Decimal("123.0")),
        ("393097812301", "EUR", Decimal("12301")),
    ],
)
def test_extract_amount_payable_and_currency(
    value: str, expected_currency: str | None, expected_decimal: Decimal | None
) -> None:
    element_string = GS1ElementString.extract(value)

    assert element_string.decimal == expected_decimal

    # Optional: If py-moneyed is installed, create Money instances
    if expected_decimal is None:
        assert element_string.money is None
    else:
        assert element_string.money is not None
        assert element_string.money.amount == expected_decimal
        assert element_string.money.currency.code == expected_currency


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("39400010", Decimal("10")),
        ("39410055", Decimal("5.5")),
    ],
)
def test_extract_percentage_discount(value: str, expected: Decimal) -> None:
    assert GS1ElementString.extract(value).decimal == expected


@pytest.mark.parametrize(
    ("value", "expected"), [("0107032069804988", "(01)07032069804988")]
)
def test_as_hri(value: str, expected: str) -> None:
    assert GS1ElementString.extract(value).as_hri() == expected


@pytest.mark.parametrize(
    ("value", "ai", "expected"),
    [
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "01",
            ["00614141123452"],
        ),
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "10",
            ["ABC123"],
        ),
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            GS1ApplicationIdentifier.extract("10"),
            ["ABC123"],
        ),
    ],
)
def test_list_filter_element_strings_by_ai(
    value: list[GS1ElementString],
    ai: str | GS1ApplicationIdentifier,
    expected: list[str],
) -> None:
    matches = GS1ElementStrings(value).filter(ai=ai)

    assert [element_string.value for element_string in matches] == expected


@pytest.mark.parametrize(
    ("value", "data_title", "expected"),
    [
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "GTIN",
            ["00614141123452"],
        ),
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "BATCH",
            ["ABC123"],
        ),
    ],
)
def test_list_filter_element_strings_by_data_title(
    value: list[GS1ElementString],
    data_title: str,
    expected: list[str],
) -> None:
    matches = GS1ElementStrings(value).filter(data_title=data_title)

    assert [element_string.value for element_string in matches] == expected


@pytest.mark.parametrize(
    ("value", "ai", "expected"),
    [
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "01",
            "00614141123452",
        ),
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "10",
            "ABC123",
        ),
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "15",
            None,
        ),
    ],
)
def test_list_get_element_strings_by_ai(
    value: list[GS1ElementString],
    ai: str,
    expected: str | None,
) -> None:
    element_string = GS1ElementStrings(value).get(ai=ai)

    if expected is None:
        assert element_string is None
    else:
        assert element_string is not None
        assert element_string.value == expected


@pytest.mark.parametrize(
    ("value", "data_title", "expected"),
    [
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "GTIN",
            "00614141123452",
        ),
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "BATCH",
            "ABC123",
        ),
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "LOT",
            "ABC123",
        ),
        (
            [
                GS1ElementString.extract("0100614141123452"),
                GS1ElementString.extract("10ABC123"),
            ],
            "BEST BY",
            None,
        ),
    ],
)
def test_list_get_element_strings_by_data_title(
    value: list[GS1ElementString],
    data_title: str,
    expected: str | None,
) -> None:
    element_string = GS1ElementStrings(value).get(data_title=data_title)

    if expected is None:
        assert element_string is None
    else:
        assert element_string is not None
        assert element_string.value == expected



"""Global Location Number (GLN).

GLNs are used to identify physical locations, digital locations, legal entities,
and organizational functions.

If you only want to parse GLNs, you can import the GLN parser directly.

    >>> from biip.gln import Gln

If parsing succeeds, it returns a `Gln` object.

    >>> gln = Gln.parse("1234567890128")
    >>> pprint(gln)
    Gln(
        value='1234567890128',
        prefix=GS1Prefix(
            value='123',
            usage='GS1 US'
        ),
        company_prefix=GS1CompanyPrefix(
            value='1234567890'
        ),
        payload='123456789012',
        check_digit=8
    )

As GLNs do not appear independently in barcodes, the GLN parser is not a part of
the top-level parser [`biip.parse()`][biip.parse]. However, if you are parsing a
barcode with GS1 element strings including a GLN, the GLN will be parsed and
validated using the [`Gln`][biip.gln.Gln] class.

    >>> import biip
    >>> gln = (
    ...     biip.parse("4101234567890128")
    ...     .gs1_message
    ...     .element_strings.get(data_title="SHIP TO")
    ...     .gln
    ... )
    >>> pprint(gln)
    Gln(
        value='1234567890128',
        prefix=GS1Prefix(
            value='123',
            usage='GS1 US'
        ),
        company_prefix=GS1CompanyPrefix(
            value='1234567890'
        ),
        payload='123456789012',
        check_digit=8
    )
"""

from __future__ import annotations

from dataclasses import dataclass

from biip import ParseConfig, ParseError
from biip.checksums import gs1_standard_check_digit
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix


@dataclass(frozen=True)
class Gln:
    """Dataclass containing a GLN."""

    value: str
    """Raw unprocessed value."""

    prefix: GS1Prefix | None
    """The [GS1 Prefix][biip.gs1_prefixes.GS1Prefix].

    Indicating what GS1 country organization that assigned code range.
    """

    company_prefix: GS1CompanyPrefix | None
    """The [GS1 Company Prefix][biip.gs1_prefixes.GS1CompanyPrefix].

    Identifying the company that issued the GLN.
    """

    payload: str
    """The actual payload.

    Including extension digit, company prefix, and item reference. Excludes the
    check digit.
    """

    check_digit: int
    """Check digit used to check if the GLN as a whole is valid."""

    @classmethod
    def parse(
        cls,
        value: str,
        *,
        config: ParseConfig | None = None,  # noqa: ARG003
    ) -> Gln:
        """Parse the given value into a [`Gln`][biip.gln.Gln] object.

        The checksum is guaranteed to be valid if a GLN object is returned.

        Args:
            value: The value to parse.
            config: Configuration options for parsing.

        Returns:
            GLN data structure with the successfully extracted data.

        Raises:
            ParseError: If the parsing fails.
        """
        value = value.strip()

        if len(value) != 13:
            msg = (
                f"Failed to parse {value!r} as GLN: "
                f"Expected 13 digits, got {len(value)}."
            )
            raise ParseError(msg)

        if not value.isdecimal():
            msg = f"Failed to parse {value!r} as GLN: Expected a numerical value."
            raise ParseError(msg)

        prefix = GS1Prefix.extract(value)
        company_prefix = GS1CompanyPrefix.extract(value)
        payload = value[:-1]
        check_digit = int(value[-1])

        calculated_check_digit = gs1_standard_check_digit(payload)
        if check_digit != calculated_check_digit:
            msg = (
                f"Invalid GLN check digit for {value!r}: "
                f"Expected {calculated_check_digit!r}, got {check_digit!r}."
            )
            raise ParseError(msg)

        return cls(
            value=value,
            prefix=prefix,
            company_prefix=company_prefix,
            payload=payload,
            check_digit=check_digit,
        )

    def as_gln(self) -> str:
        """Format as a GLN."""
        return self.value



from decimal import Decimal

import pytest
from moneyed import Money

from biip import ParseConfig, ParseError
from biip.gs1_prefixes import GS1Prefix
from biip.gtin import Gtin, GtinFormat
from biip.rcn import Rcn, RcnRegion, RcnUsage


@pytest.mark.parametrize(
    ("rcn_region", "value", "weight", "price", "money"),
    [
        # NOTE: These examples are constructed from a template. This should be
        # extended with actual examples from either specifications or real
        # Baltic products.
        #
        # Estonia
        (RcnRegion.ESTONIA, "2311111112345", Decimal("1.234"), None, None),
        (RcnRegion.ESTONIA, "2411111112342", Decimal("12.34"), None, None),
        (RcnRegion.ESTONIA, "2511111112349", Decimal("123.4"), None, None),
        (RcnRegion.ESTONIA, "2911111111111", None, None, None),
        # Latvia
        (RcnRegion.LATVIA, "2311111112345", Decimal("1.234"), None, None),
        (RcnRegion.LATVIA, "2411111112342", Decimal("12.34"), None, None),
        (RcnRegion.LATVIA, "2511111112349", Decimal("123.4"), None, None),
        (RcnRegion.LATVIA, "2911111111111", None, None, None),
        # Lithuania
        (RcnRegion.LITHUANIA, "2311111112345", Decimal("1.234"), None, None),
        (RcnRegion.LITHUANIA, "2411111112342", Decimal("12.34"), None, None),
        (RcnRegion.LITHUANIA, "2511111112349", Decimal("123.4"), None, None),
        (RcnRegion.LITHUANIA, "2911111111111", None, None, None),
    ],
)
def test_region_baltics(
    rcn_region: RcnRegion,
    value: str,
    weight: Decimal | None,
    price: Decimal | None,
    money: Money | None,
) -> None:
    # The three Baltic countries share the same rules and allocation pool.
    #
    # References:
    #   https://gs1lv.org/img/upload/ENG.Variable%20measure_in_Latvia.pdf

    rcn = Gtin.parse(value, config=ParseConfig(rcn_region=rcn_region))

    assert isinstance(rcn, Rcn)
    assert rcn.region == rcn_region
    assert rcn.weight == weight
    assert rcn.price == price
    assert rcn.money == money


@pytest.mark.parametrize(
    ("value", "weight", "price", "money"),
    [
        ("2135040039753", None, Decimal("39.75"), Money("39.75", "DKK")),
        ("2235040039750", None, Decimal("39.75"), Money("39.75", "DKK")),
        ("2335040039757", None, Decimal("39.75"), Money("39.75", "DKK")),
        ("2435040039754", None, Decimal("39.75"), Money("39.75", "DKK")),
        ("2623570001505", Decimal("0.150"), None, None),
        ("2712341002251", Decimal("0.225"), None, None),
        ("2823570001509", Decimal("0.150"), None, None),
        ("2911111111111", None, None, None),
    ],
)
def test_region_denmark(
    value: str,
    weight: Decimal | None,
    price: Decimal | None,
    money: Money | None,
) -> None:
    # References:
    #   https://www.gs1.dk/om-gs1/overblik-over-gs1-standarder/gtin-13-pris
    #   https://www.gs1.dk/om-gs1/overblik-over-gs1-standarder/gtin-13-vaegt

    rcn = Gtin.parse(value, config=ParseConfig(rcn_region=RcnRegion.DENMARK))

    assert isinstance(rcn, Rcn)
    assert rcn.region == RcnRegion.DENMARK
    assert rcn.weight == weight
    assert rcn.price == price
    assert rcn.money == money


@pytest.mark.parametrize(
    ("value", "weight", "price", "money"),
    [
        ("2388060112344", Decimal("1.234"), None, None),
        ("2488060112341", Decimal("12.34"), None, None),
        ("2588060112348", Decimal("123.4"), None, None),
        ("2911111111111", None, None, None),
    ],
)
def test_region_finland(
    value: str,
    weight: Decimal | None,
    price: Decimal | None,
    money: Money | None,
) -> None:
    # References:
    #   https://gs1.fi/en/instructions/gs1-company-prefix/how-identify-product-gtin

    rcn = Gtin.parse(value, config=ParseConfig(rcn_region=RcnRegion.FINLAND))

    assert isinstance(rcn, Rcn)
    assert rcn.region == RcnRegion.FINLAND
    assert rcn.weight == weight
    assert rcn.price == price
    assert rcn.money == money


@pytest.mark.parametrize(
    ("value", "weight", "count", "price", "money"),
    [
        # Money
        ("2211114002394", None, None, Decimal("2.39"), Money("2.39", "EUR")),
        ("2330714002396", None, None, Decimal("2.39"), Money("2.39", "EUR")),
        # Count
        ("2511119000129", None, 12, None, None),
        ("2630719000121", None, 12, None, None),
        # Weight
        ("2811111068708", Decimal("6.870"), None, None, None),
        ("2930711068700", Decimal("6.870"), None, None, None),
        # GTIN-14 works exactly like GTIN-13
        ("02211114002394", None, None, Decimal("2.39"), Money("2.39", "EUR")),
        ("02511119000129", None, 12, None, None),
        ("02811111068708", Decimal("6.870"), None, None, None),
    ],
)
def test_region_germany(
    value: str,
    weight: Decimal | None,
    count: int | None,
    price: Decimal | None,
    money: Money | None,
) -> None:
    # References:
    #   https://www.gs1-germany.de/fileadmin/gs1/fachpublikationen/globale-artikelnummer-gtin-in-der-anwendung.pdf
    #   https://san.gs1-germany.de/SAN-4-Konzept

    rcn = Gtin.parse(value, config=ParseConfig(rcn_region=RcnRegion.GERMANY))

    assert isinstance(rcn, Rcn)
    assert rcn.region == RcnRegion.GERMANY
    assert rcn.weight == weight
    assert rcn.count == count
    assert rcn.price == price
    assert rcn.money == money


def test_region_germany_fails_with_invalid_variable_measure_check_digit() -> None:
    # The digit 8 in the value below is the variable measure check digit. The
    # correct value is 9.

    with pytest.raises(ParseError) as exc_info:
        Gtin.parse("2511118000120", config=ParseConfig(rcn_region=RcnRegion.GERMANY))

    assert str(exc_info.value) == (
        "Invalid check digit for variable measure value '00012' "
        "in RCN '2511118000120': Expected 9, got 8."
    )


def test_region_germany_when_not_verifying_invalid_check_digit() -> None:
    # The digit 8 in the value below is the variable measure check digit. The
    # correct value is 9.

    rcn = Gtin.parse(
        "2511118000120",
        config=ParseConfig(
            rcn_region=RcnRegion.GERMANY,
            rcn_verify_variable_measure=False,
        ),
    )

    assert isinstance(rcn, Rcn)
    assert rcn == Rcn(
        value="2511118000120",
        format=GtinFormat.GTIN_13,
        prefix=GS1Prefix.extract("251"),
        company_prefix=None,
        payload="251111800012",
        check_digit=0,
        region=RcnRegion.GERMANY,
        usage=RcnUsage.GEOGRAPHICAL,
        count=12,
    )


@pytest.mark.parametrize(
    ("value", "weight", "price", "money"),
    [
        # NOTE: These examples are constructed from a template. This should be
        # extended with actual examples from either specifications or real
        # British products.
        ("2011122912346", None, Decimal("12.34"), Money("12.34", "GBP")),
        ("2911111111111", None, None, None),
    ],
)
def test_region_great_britain(
    value: str,
    weight: Decimal | None,
    price: Decimal | None,
    money: Money | None,
) -> None:
    # References:
    #   https://www.gs1uk.org/how-to-barcode-variable-measure-items

    rcn = Gtin.parse(value, config=ParseConfig(rcn_region=RcnRegion.GREAT_BRITAIN))

    assert isinstance(rcn, Rcn)
    assert rcn.region == RcnRegion.GREAT_BRITAIN
    assert rcn.weight == weight
    assert rcn.price == price
    assert rcn.money == money


def test_region_great_britain_fails_with_invalid_price_check_digit() -> None:
    # The digit 8 in the value below is the price check digit. The correct value is 9.

    with pytest.raises(ParseError) as exc_info:
        Gtin.parse(
            "2011122812349",
            config=ParseConfig(rcn_region=RcnRegion.GREAT_BRITAIN),
        )

    assert str(exc_info.value) == (
        "Invalid check digit for variable measure value '1234' in RCN '2011122812349': "
        "Expected 9, got 8."
    )


def test_region_great_britain_when_not_verifying_invalid_check_digit() -> None:
    # The digit 8 in the value below is the price check digit. The correct value is 9.

    rcn = Gtin.parse(
        "2011122812349",
        config=ParseConfig(
            rcn_region=RcnRegion.GREAT_BRITAIN,
            rcn_verify_variable_measure=False,
        ),
    )

    assert isinstance(rcn, Rcn)
    assert rcn == Rcn(
        value="2011122812349",
        format=GtinFormat.GTIN_13,
        prefix=GS1Prefix.extract("201"),
        company_prefix=None,
        payload="201112281234",
        check_digit=9,
        region=RcnRegion.GREAT_BRITAIN,
        usage=RcnUsage.GEOGRAPHICAL,
        price=Decimal("12.34"),
        money=Money("12.34", "GBP"),
    )


@pytest.mark.parametrize(
    ("value", "weight", "price", "money"),
    [
        ("2302148210869", Decimal("1.086"), None, None),  # Norvegia 1kg
        ("2368091402263", Decimal("0.226"), None, None),  # Stange kyllingbryst
        ("2911111111111", None, None, None),
    ],
)
def test_region_norway(
    value: str,
    weight: Decimal | None,
    price: Decimal | None,
    money: Money | None,
) -> None:
    # References: TODO: Find specification.

    rcn = Gtin.parse(value, config=ParseConfig(rcn_region=RcnRegion.NORWAY))

    assert isinstance(rcn, Rcn)
    assert rcn.region == RcnRegion.NORWAY
    assert rcn.weight == weight
    assert rcn.price == price
    assert rcn.money == money


@pytest.mark.parametrize(
    ("value", "weight", "price", "money"),
    [
        ("2088060112343", None, Decimal("12.34"), Money("12.34", "SEK")),
        ("2188060112340", None, Decimal("123.4"), Money("123.4", "SEK")),
        ("2288060112347", None, Decimal("1234"), Money("1234", "SEK")),
        ("2388060112344", Decimal("1.234"), None, None),
        ("2488060112341", Decimal("12.34"), None, None),
        ("2588060112348", Decimal("123.4"), None, None),
        ("2911111111111", None, None, None),
    ],
)
def test_region_sweden(
    value: str,
    weight: Decimal | None,
    price: Decimal | None,
    money: Money | None,
) -> None:
    # References:
    #   https://www.gs1.se/en/our-standards/Identify/variable-weight-number1/

    rcn = Gtin.parse(value, config=ParseConfig(rcn_region=RcnRegion.SWEDEN))

    assert isinstance(rcn, Rcn)
    assert rcn.region == RcnRegion.SWEDEN
    assert rcn.weight == weight
    assert rcn.price == price
    assert rcn.money == money



import datetime as dt
from collections.abc import Iterable
from decimal import Decimal

import pytest

from biip import ParseConfig, ParseError
from biip.gs1_application_identifiers import GS1ApplicationIdentifier
from biip.gs1_element_strings import GS1ElementString, GS1ElementStrings
from biip.gs1_messages import GS1Message
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix
from biip.gtin import Gtin, GtinFormat
from biip.rcn import Rcn, RcnRegion


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "010703206980498815210526100329",
            GS1Message(
                value="010703206980498815210526100329",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("01"),
                            value="07032069804988",
                            pattern_groups=["07032069804988"],
                            gtin=Gtin(
                                value="07032069804988",
                                format=GtinFormat.GTIN_13,
                                prefix=GS1Prefix(value="703", usage="GS1 Norway"),
                                company_prefix=GS1CompanyPrefix(value="703206"),
                                payload="703206980498",
                                check_digit=8,
                            ),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("15"),
                            value="210526",
                            pattern_groups=["210526"],
                            date=dt.date(2021, 5, 26),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("10"),
                            value="0329",
                            pattern_groups=["0329"],
                        ),
                    ]
                ),
            ),
        ),
        (
            "800307071324001022085952",
            GS1Message(
                value="800307071324001022085952",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("8003"),
                            value="07071324001022085952",
                            pattern_groups=["0", "7071324001022", "085952"],
                        )
                    ]
                ),
            ),
        ),
    ],
)
def test_parse(value: str, expected: GS1Message) -> None:
    assert GS1Message.parse(value) == expected


@pytest.mark.parametrize(
    ("value", "separator_chars", "expected_hri"),
    [
        (
            # Variable-length lot number field last, all OK.
            "010703206980498815210526100329",
            ("\x1d",),
            "(01)07032069804988(15)210526(10)0329",
        ),
        (
            # Variable-length lot number field in the middle, consuming the
            # best before date field at the end.
            "010703206980498810032915210526",
            ("\x1d",),
            "(01)07032069804988(10)032915210526",
        ),
        (
            # Variable-length lot number field in the middle, end marked with
            # default separator character.
            "0107032069804988100329\x1d15210526",
            ("\x1d",),
            "(01)07032069804988(10)0329(15)210526",
        ),
        (
            # Variable-length lot number field in the middle, end marked with
            # custom separator character.
            "0107032069804988100329|15210526",
            ("|",),
            "(01)07032069804988(10)0329(15)210526",
        ),
        (
            # Unrealistic corner case just to exercise the code:
            # Two variable-length fields marked with different separators
            "0107032069804988100329|2112345\x1d15210526",
            ("|", "\x1d"),
            "(01)07032069804988(10)0329(21)12345(15)210526",
        ),
    ],
)
def test_parse_with_separator_char(
    value: str, separator_chars: Iterable[str], expected_hri: str
) -> None:
    assert (
        GS1Message.parse(
            value,
            config=ParseConfig(separator_chars=separator_chars),
        ).as_hri()
        == expected_hri
    )


@pytest.mark.parametrize(
    ("value", "config", "expected"),
    [
        (
            # Invalid date "000000"
            "15000000",
            ParseConfig(gs1_element_strings_verify_date=False),
            GS1Message(
                value="15000000",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("15"),
                            value="000000",
                            pattern_groups=["000000"],
                            date=None,
                        )
                    ]
                ),
            ),
        ),
    ],
)
def test_parse_with_invalid_element_values(
    value: str,
    config: ParseConfig,
    expected: GS1Message,
) -> None:
    assert GS1Message.parse(value, config=config) == expected


def test_parse_with_too_long_separator_char_fails() -> None:
    with pytest.raises(
        ValueError,
        match=(
            r"^All separator characters must be exactly 1 character long, "
            r"got \['--'\].$"
        ),
    ):
        GS1Message.parse(
            "10222--15210526",
            config=ParseConfig(separator_chars=["--"]),
        )


@pytest.mark.parametrize(
    ("value", "error"),
    [
        # 10 = AI for BATCH/LOT
        # 222... = Max length BATCH/LOT
        # aaa = Superflous data
        (
            "1022222222222222222222aaa",
            "Failed to get GS1 Application Identifier from 'aaa'.",
        ),
        # Too short to match optional time group (as this is really a GTIN-13)
        ("701103020185", "Failed to get GS1 Application Identifier from '85'."),
    ],
)
def test_parse_fails_if_unparsed_data_left(value: str, error: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        GS1Message.parse(value)

    assert str(exc_info.value) == error


@pytest.mark.parametrize(
    ("value", "expected_hri"),
    [
        # A separator at the end of the string is never needed.
        ("15210526100329\x1d", "(15)210526(10)0329"),
        # 15... is a fixed length date, and should not end with a separator.
        ("15210526\x1d100329", "(15)210526(10)0329"),
        # Double separator chars are ignored.
        (
            "0107032069804988100329\x1d\x1d15210526",
            "(01)07032069804988(10)0329(15)210526",
        ),
    ],
)
def test_parse_ignores_redundant_separator_chars(value: str, expected_hri: str) -> None:
    assert GS1Message.parse(value).as_hri() == expected_hri


def test_parse_strips_surrounding_whitespace() -> None:
    message = GS1Message.parse("  \t 800307071324001022085952 \n  ")

    assert message.value == "800307071324001022085952"


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            GS1ElementStrings(
                [
                    GS1ElementString(
                        ai=GS1ApplicationIdentifier.extract("17"),
                        value="221231",
                        pattern_groups=["221231"],
                        date=dt.date(2022, 12, 31),
                    )
                ]
            ),
            GS1Message(
                value="17221231",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("17"),
                            value="221231",
                            pattern_groups=["221231"],
                            date=dt.date(2022, 12, 31),
                        )
                    ]
                ),
            ),
        ),
        (
            GS1ElementStrings(
                [
                    GS1ElementString(
                        ai=GS1ApplicationIdentifier.extract("10"),
                        value="123",
                        pattern_groups=["123"],
                    ),
                    GS1ElementString(
                        ai=GS1ApplicationIdentifier.extract("17"),
                        value="221231",
                        pattern_groups=["221231"],
                        date=dt.date(2022, 12, 31),
                    ),
                ]
            ),
            GS1Message(
                value="10123\x1d17221231",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("10"),
                            value="123",
                            pattern_groups=["123"],
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("17"),
                            value="221231",
                            pattern_groups=["221231"],
                            date=dt.date(2022, 12, 31),
                        ),
                    ]
                ),
            ),
        ),
    ],
)
def test_from_element_strings(
    value: GS1ElementStrings,
    expected: GS1Message,
) -> None:
    assert GS1Message.from_element_strings(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "(17)221231",
            GS1Message(
                value="17221231",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("17"),
                            value="221231",
                            pattern_groups=["221231"],
                            date=dt.date(2022, 12, 31),
                        )
                    ]
                ),
            ),
        ),
        (
            "(10)123(17)221231",
            GS1Message(
                value="10123\x1d17221231",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("10"),
                            value="123",
                            pattern_groups=["123"],
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("17"),
                            value="221231",
                            pattern_groups=["221231"],
                            date=dt.date(2022, 12, 31),
                        ),
                    ]
                ),
            ),
        ),
        (
            "(17)221231(10)123",
            GS1Message(
                value="1722123110123",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("17"),
                            value="221231",
                            pattern_groups=["221231"],
                            date=dt.date(2022, 12, 31),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("10"),
                            value="123",
                            pattern_groups=["123"],
                        ),
                    ]
                ),
            ),
        ),
    ],
)
def test_parse_hri(value: str, expected: GS1Message) -> None:
    assert GS1Message.parse_hri(value) == expected


def test_parse_hri_with_gtin_with_variable_weight() -> None:
    result = GS1Message.parse_hri(
        "(01)02302148210869",
        config=ParseConfig(rcn_region=RcnRegion.NORWAY),
    )

    gs1_gtin = result.element_strings.get(ai="01")
    assert gs1_gtin
    gtin = gs1_gtin.gtin
    assert isinstance(gtin, Rcn)
    assert gtin.weight == Decimal("1.086")


def test_parse_hri_strips_surrounding_whitespace() -> None:
    message = GS1Message.parse_hri("  \t (17)221231 \n  ")

    assert message.value == "17221231"


@pytest.mark.parametrize(
    "value",
    [
        "",  # Empty string
        "17221231",  # Valid data, but no parenthesis
        "aaa(17)221231",  # Valid data, but extra data in front
    ],
)
def test_parse_hri_fails_if_not_starting_with_parenthesis(value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        GS1Message.parse_hri(value)

    assert str(exc_info.value) == (
        f"Expected HRI string {value!r} to start with a parenthesis."
    )


@pytest.mark.parametrize(
    "value",
    [
        "(15)",  # Valid start of string, but no data
    ],
)
def test_parse_hri_fails_if_no_pattern_matches(value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        GS1Message.parse_hri(value)

    assert str(exc_info.value) == (
        "Could not find any GS1 Application Identifiers in "
        f"{value!r}. Expected format: '(AI)DATA(AI)DATA'."
    )


@pytest.mark.parametrize(
    "value",
    [
        "(1)15210526",
    ],
)
def test_parse_hri_fails_if_ai_is_unknown(value: str) -> None:
    with pytest.raises(ParseError) as exc_info:
        GS1Message.parse_hri(value)

    assert str(exc_info.value) == (
        "Unknown GS1 Application Identifier '1' in '(1)15210526'."
    )


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "010703206980498815210526100329",
            "(01)07032069804988(15)210526(10)0329",
        )
    ],
)
def test_as_hri(value: str, expected: str) -> None:
    assert GS1Message.parse(value).as_hri() == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "010703206980498815210526100329",
            "https://id.gs1.org/01/07032069804988/10/0329?15=210526",
        )
    ],
)
def test_as_gs1_web_uri(value: str, expected: str) -> None:
    assert GS1Message.parse(value).as_gs1_web_uri().value == expected


@pytest.mark.parametrize(
    ("value", "ai", "expected"),
    [
        ("010703206980498815210526100329", "01", ["07032069804988"]),
        ("010703206980498815210526100329", "15", ["210526"]),
        ("010703206980498815210526100329", "37", []),
        ("7230EM123\x1d7231EM456\x1d7232EM789", "7231", ["EM456"]),
        (
            "7230EM123\x1d7231EM456\x1d7232EM789",
            "723",
            ["EM123", "EM456", "EM789"],
        ),
    ],
)
def test_filter_element_strings_by_ai(value: str, ai: str, expected: list[str]) -> None:
    matches = GS1Message.parse(value).element_strings.filter(ai=ai)

    assert [element_string.value for element_string in matches] == expected


@pytest.mark.parametrize(
    ("value", "data_title", "expected"),
    [
        ("010703206980498815210526100329", "GTIN", ["07032069804988"]),
        ("010703206980498815210526100329", "BEST BY", ["210526"]),
        ("010703206980498815210526100329", "COUNT", []),
        (
            "7230EM123\x1d7231EM456\x1d7232EM789",
            "CERT",
            ["EM123", "EM456", "EM789"],
        ),
    ],
)
def test_filter_element_strings_by_data_title(
    value: str, data_title: str, expected: list[str]
) -> None:
    matches = GS1Message.parse(value).element_strings.filter(data_title=data_title)

    assert [element_string.value for element_string in matches] == expected


@pytest.mark.parametrize(
    ("value", "ai", "expected"),
    [
        ("010703206980498815210526100329", "01", "07032069804988"),
        ("010703206980498815210526100329", "15", "210526"),
        ("010703206980498815210526100329", "10", "0329"),
        ("010703206980498815210526100329", "37", None),
        ("7230EM123\x1d7231EM456\x1d7232EM789", "7231", "EM456"),
        ("7230EM123\x1d7231EM456\x1d7232EM789", "723", "EM123"),
    ],
)
def test_get_element_string_by_ai(value: str, ai: str, expected: str | None) -> None:
    element_string = GS1Message.parse(value).element_strings.get(ai=ai)

    if expected is None:
        assert element_string is None
    else:
        assert element_string is not None
        assert element_string.value == expected


@pytest.mark.parametrize(
    ("value", "data_title", "expected"),
    [
        ("010703206980498815210526100329", "GTIN", "07032069804988"),
        ("010703206980498815210526100329", "BEST BY", "210526"),
        ("010703206980498815210526100329", "BATCH", "0329"),
        ("010703206980498815210526100329", "COUNT", None),
        ("7230EM123\x1d7231EM456\x1d7232EM789", "CERT # 1", "EM456"),
        ("7230EM123\x1d7231EM456\x1d7232EM789", "CERT", "EM123"),
    ],
)
def test_get_element_string_by_data_title(
    value: str, data_title: str, expected: str | None
) -> None:
    element_string = GS1Message.parse(value).element_strings.get(data_title=data_title)

    if expected is None:
        assert element_string is None
    else:
        assert element_string is not None
        assert element_string.value == expected


def test_filter_element_strings_by_ai_instance() -> None:
    ai = GS1ApplicationIdentifier.extract("01")
    msg = GS1Message.parse("010703206980498815210526100329")

    element_string = msg.element_strings.get(ai=ai)

    assert element_string is not None
    assert element_string.value == "07032069804988"



import datetime as dt

import pytest

from biip import ParseConfig, ParseError
from biip.gs1_application_identifiers import GS1ApplicationIdentifier
from biip.gs1_element_strings import GS1ElementString, GS1ElementStrings
from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix
from biip.gs1_web_uris import GS1WebURI
from biip.gtin import Gtin, GtinFormat
from biip.sscc import Sscc


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            # GTIN-12 by AI number
            "https://id.gs1.org/01/614141123452",
            GS1WebURI(
                value="https://id.gs1.org/01/614141123452",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("01"),
                            value="00614141123452",
                            pattern_groups=["00614141123452"],
                            gtin=Gtin(
                                value="00614141123452",
                                format=GtinFormat.GTIN_12,
                                prefix=GS1Prefix(value="061", usage="GS1 US"),
                                company_prefix=GS1CompanyPrefix(value="0614141"),
                                payload="61414112345",
                                check_digit=2,
                            ),
                        ),
                    ]
                ),
            ),
        ),
        (
            # GTIN-12 with custom domain, path prefix, and AI alias
            "https://example.com/foo/gtin/614141123452",
            GS1WebURI(
                value="https://example.com/foo/gtin/614141123452",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("01"),
                            value="00614141123452",
                            pattern_groups=["00614141123452"],
                            gtin=Gtin(
                                value="00614141123452",
                                format=GtinFormat.GTIN_12,
                                prefix=GS1Prefix(value="061", usage="GS1 US"),
                                company_prefix=GS1CompanyPrefix(value="0614141"),
                                payload="61414112345",
                                check_digit=2,
                            ),
                        ),
                    ]
                ),
            ),
        ),
        (
            # GTIN-12 and batch/lot number by AI number
            "https://id.gs1.org/01/614141123452/10/ABC123",
            GS1WebURI(
                value="https://id.gs1.org/01/614141123452/10/ABC123",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("01"),
                            value="00614141123452",
                            pattern_groups=["00614141123452"],
                            gtin=Gtin(
                                value="00614141123452",
                                format=GtinFormat.GTIN_12,
                                prefix=GS1Prefix(value="061", usage="GS1 US"),
                                company_prefix=GS1CompanyPrefix(value="0614141"),
                                payload="61414112345",
                                check_digit=2,
                            ),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("10"),
                            value="ABC123",
                            pattern_groups=["ABC123"],
                        ),
                    ]
                ),
            ),
        ),
        (
            # GTIN-12, CPV, and batch/lot number by AI alias
            "https://id.gs1.org/gtin/614141123452/cpv/2A/lot/ABC123",
            GS1WebURI(
                value="https://id.gs1.org/gtin/614141123452/cpv/2A/lot/ABC123",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("01"),
                            value="00614141123452",
                            pattern_groups=["00614141123452"],
                            gtin=Gtin(
                                value="00614141123452",
                                format=GtinFormat.GTIN_12,
                                prefix=GS1Prefix(value="061", usage="GS1 US"),
                                company_prefix=GS1CompanyPrefix(value="0614141"),
                                payload="61414112345",
                                check_digit=2,
                            ),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("22"),
                            value="2A",
                            pattern_groups=["2A"],
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("10"),
                            value="ABC123",
                            pattern_groups=["ABC123"],
                        ),
                    ]
                ),
            ),
        ),
        (
            # GTIN-12 with extra query parameter that is ignored
            "https://id.gs1.org/01/614141123452?foo=bar",
            GS1WebURI(
                value="https://id.gs1.org/01/614141123452?foo=bar",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("01"),
                            value="00614141123452",
                            pattern_groups=["00614141123452"],
                            gtin=Gtin(
                                value="00614141123452",
                                format=GtinFormat.GTIN_12,
                                prefix=GS1Prefix(value="061", usage="GS1 US"),
                                company_prefix=GS1CompanyPrefix(value="0614141"),
                                payload="61414112345",
                                check_digit=2,
                            ),
                        ),
                    ]
                ),
            ),
        ),
        (
            # SSCC with content, count, and batch/lot number
            "https://id.gs1.org/sscc/106141412345678908?02=00614141123452&37=25&10=ABC123",
            GS1WebURI(
                value="https://id.gs1.org/sscc/106141412345678908?02=00614141123452&37=25&10=ABC123",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("00"),
                            value="106141412345678908",
                            pattern_groups=["106141412345678908"],
                            sscc=Sscc(
                                value="106141412345678908",
                                prefix=GS1Prefix(value="061", usage="GS1 US"),
                                company_prefix=GS1CompanyPrefix(value="0614141"),
                                extension_digit=1,
                                payload="10614141234567890",
                                check_digit=8,
                            ),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("02"),
                            value="00614141123452",
                            pattern_groups=["00614141123452"],
                            gln=None,
                            gln_error=None,
                            gtin=Gtin(
                                value="00614141123452",
                                format=GtinFormat.GTIN_12,
                                prefix=GS1Prefix(value="061", usage="GS1 US"),
                                company_prefix=GS1CompanyPrefix(value="0614141"),
                                payload="61414112345",
                                check_digit=2,
                                packaging_level=None,
                            ),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("37"),
                            value="25",
                            pattern_groups=["25"],
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("10"),
                            value="ABC123",
                            pattern_groups=["ABC123"],
                        ),
                    ]
                ),
            ),
        ),
    ],
)
def test_parse(value: str, expected: GS1WebURI) -> None:
    assert GS1WebURI.parse(value) == expected


@pytest.mark.parametrize(
    ("value", "config", "expected"),
    [
        (
            # Invalid date "000000"
            "https://id.gs1.org/gtin/614141123452?15=000000",
            ParseConfig(gs1_element_strings_verify_date=False),
            GS1WebURI(
                value="https://id.gs1.org/gtin/614141123452?15=000000",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("01"),
                            value="00614141123452",
                            pattern_groups=["00614141123452"],
                            gtin=Gtin(
                                value="00614141123452",
                                format=GtinFormat.GTIN_12,
                                prefix=GS1Prefix(value="061", usage="GS1 US"),
                                company_prefix=GS1CompanyPrefix(value="0614141"),
                                payload="61414112345",
                                check_digit=2,
                            ),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("15"),
                            value="000000",
                            pattern_groups=["000000"],
                            date=None,
                        ),
                    ]
                ),
            ),
        ),
    ],
)
def test_parse_with_invalid_element_values(
    value: str,
    config: ParseConfig,
    expected: GS1WebURI,
) -> None:
    assert GS1WebURI.parse(value, config=config) == expected


@pytest.mark.parametrize(
    ("value", "error"),
    [
        (
            # Non-URI are not allowed
            "614141123452",
            r"^Expected URI, got '614141123452'.$",
        ),
        (
            # Non-HTTP URIs are not allowed
            "mailto:alice@example.com",
            r"^Expected URI scheme to be 'http' or 'https', got 'mailto'.$",
        ),
        (
            # A primary identifier in the path is required
            "http://example.com/foo/123",
            r"^Expected a primary identifier in the path, got '/foo/123'.$",
        ),
        (
            # The path must contain an even number of segments
            "https://example.com/foo/gtin/614141123452/123",
            r"^Expected even number of path segments, got '/gtin/614141123452/123'.$",
        ),
        (
            # "exp" is not a valid qualifier for GTIN
            "https://id.gs1.org/gtin/614141123452/exp/20250324",
            r"^Expected one of 22/cpv/10/lot/21/ser as qualifier, got 'exp'.$",
        ),
        (
            # "cpv" is a valid qualifier, but not after "lot"
            "https://id.gs1.org/gtin/614141123452/lot/ABC123/cpv/123456789",
            r"^Expected one of 21/ser as qualifier, got 'cpv'.$",
        ),
        (
            # "lot" is a valid qualifier, but not after "ser"
            "https://id.gs1.org/gtin/01234567890128/ser/12345XYZ/lot/ABC123",
            r"^Did not expect a qualifier, got 'lot'.$",
        ),
        (
            # "lot" is not a valid qualifier for SSCC
            "https://id.gs1.org/sscc/106141412345678908/lot/ABC123",
            r"^Did not expect a qualifier, got 'lot'.$",
        ),
    ],
)
def test_parse_error(value: str, error: str) -> None:
    with pytest.raises(ParseError, match=error):
        GS1WebURI.parse(value)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            GS1ElementStrings(
                [
                    GS1ElementString(
                        ai=GS1ApplicationIdentifier.extract("01"),
                        value="07032069804988",
                        pattern_groups=["07032069804988"],
                        gtin=Gtin(
                            value="07032069804988",
                            format=GtinFormat.GTIN_13,
                            prefix=GS1Prefix(value="703", usage="GS1 Norway"),
                            company_prefix=GS1CompanyPrefix(value="703206"),
                            payload="703206980498",
                            check_digit=8,
                        ),
                    ),
                    GS1ElementString(
                        ai=GS1ApplicationIdentifier.extract("15"),
                        value="210526",
                        pattern_groups=["210526"],
                        date=dt.date(2021, 5, 26),
                    ),
                    GS1ElementString(
                        ai=GS1ApplicationIdentifier.extract("10"),
                        value="0329",
                        pattern_groups=["0329"],
                    ),
                ]
            ),
            GS1WebURI(
                value="https://id.gs1.org/01/07032069804988/10/0329?15=210526",
                element_strings=GS1ElementStrings(
                    [
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("01"),
                            value="07032069804988",
                            pattern_groups=["07032069804988"],
                            gtin=Gtin(
                                value="07032069804988",
                                format=GtinFormat.GTIN_13,
                                prefix=GS1Prefix(value="703", usage="GS1 Norway"),
                                company_prefix=GS1CompanyPrefix(value="703206"),
                                payload="703206980498",
                                check_digit=8,
                            ),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("15"),
                            value="210526",
                            pattern_groups=["210526"],
                            date=dt.date(2021, 5, 26),
                        ),
                        GS1ElementString(
                            ai=GS1ApplicationIdentifier.extract("10"),
                            value="0329",
                            pattern_groups=["0329"],
                        ),
                    ]
                ),
            ),
        ),
    ],
)
def test_from_element_strings(value: GS1ElementStrings, expected: GS1WebURI) -> None:
    assert GS1WebURI.from_element_strings(value) == expected


@pytest.mark.parametrize(
    ("value", "domain", "prefix", "short_names", "expected"),
    [
        (
            # Defaults creates a canonical URI
            "https://example.com/01/614141123452",
            None,
            None,
            False,
            "https://id.gs1.org/01/00614141123452",
        ),
        (
            # Custom domain
            "https://example.com/01/614141123452",
            "brand.example.net",
            None,
            False,
            "https://brand.example.net/01/00614141123452",
        ),
        (
            # Custom domain with prefix
            "https://example.com/gtin/614141123452",
            "brand.example.net",
            "prefix",
            False,
            "https://brand.example.net/prefix/01/00614141123452",
        ),
        (
            # Short names instead of AIs in path
            "https://id.gs1.org/01/614141123452/22/2A/10/ABC123",
            "example.com",
            "products",
            True,
            "https://example.com/products/gtin/00614141123452/cpv/2A/lot/ABC123",
        ),
        (
            # Values in params always use AI, not short names
            "https://example.com/sscc/106141412345678908?02=00614141123452&37=25&10=ABC123&foo=bar",
            "brand.example.net",
            None,
            True,
            "https://brand.example.net/sscc/106141412345678908?02=00614141123452&37=25&10=ABC123",
        ),
    ],
)
def test_as_uri(
    *,
    value: str,
    domain: str | None,
    prefix: str | None,
    short_names: bool,
    expected: str,
) -> None:
    assert (
        GS1WebURI.parse(value).as_uri(
            domain=domain,
            prefix=prefix,
            short_names=short_names,
        )
        == expected
    )


def test_as_uri_errors() -> None:
    with pytest.raises(ValueError, match="Prefix must not start with '/'"):
        GS1WebURI.parse("https://example.com/gtin/614141123452").as_uri(
            prefix="/prefix"
        )
    with pytest.raises(ValueError, match="Prefix must not end with '/'"):
        GS1WebURI.parse("https://example.com/gtin/614141123452").as_uri(
            prefix="prefix/"
        )


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "https://example.com/01/614141123452",
            "https://id.gs1.org/01/00614141123452",
        ),
        (
            "https://example.com/gtin/614141123452",
            "https://id.gs1.org/01/00614141123452",
        ),
        (
            "https://example.com/gtin/614141123452/cpv/2A/lot/ABC123",
            "https://id.gs1.org/01/00614141123452/22/2A/10/ABC123",
        ),
        (
            "https://example.com/sscc/106141412345678908?02=00614141123452&37=25&10=ABC123&foo=bar",
            "https://id.gs1.org/00/106141412345678908?02=00614141123452&37=25&10=ABC123",
        ),
    ],
)
def test_as_canonical_uri(value: str, expected: str) -> None:
    assert GS1WebURI.parse(value).as_canonical_uri() == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (
            "https://example.com/01/614141123452",
            "(01)00614141123452",
        ),
        (
            "https://example.com/gtin/614141123452",
            "(01)00614141123452",
        ),
        (
            "https://example.com/gtin/614141123452/cpv/2A/lot/ABC123",
            "(01)00614141123452(22)2A(10)ABC123",
        ),
        (
            "https://example.com/sscc/106141412345678908?02=00614141123452&37=25&10=ABC123&foo=bar",
            "(00)106141412345678908(02)00614141123452(37)25(10)ABC123",
        ),
    ],
)
def test_as_gs1_message(value: str, expected: str) -> None:
    assert GS1WebURI.parse(value).as_gs1_message().as_hri() == expected


@pytest.mark.parametrize(
    ("value", "ai", "expected"),
    [
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "01", ["00614141123452"]),
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "10", ["ABC123"]),
    ],
)
def test_filter_element_strings_by_ai(value: str, ai: str, expected: list[str]) -> None:
    matches = GS1WebURI.parse(value).element_strings.filter(ai=ai)

    assert [element_string.value for element_string in matches] == expected


@pytest.mark.parametrize(
    ("value", "data_title", "expected"),
    [
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "GTIN", ["00614141123452"]),
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "BATCH", ["ABC123"]),
    ],
)
def test_filter_element_strings_by_data_title(
    value: str, data_title: str, expected: list[str]
) -> None:
    matches = GS1WebURI.parse(value).element_strings.filter(data_title=data_title)

    assert [element_string.value for element_string in matches] == expected


@pytest.mark.parametrize(
    ("value", "ai", "expected"),
    [
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "01", "00614141123452"),
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "10", "ABC123"),
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "15", None),
    ],
)
def test_get_element_strings_by_ai(value: str, ai: str, expected: str | None) -> None:
    element_string = GS1WebURI.parse(value).element_strings.get(ai=ai)

    if expected is None:
        assert element_string is None
    else:
        assert element_string is not None
        assert element_string.value == expected


@pytest.mark.parametrize(
    ("value", "data_title", "expected"),
    [
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "GTIN", "00614141123452"),
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "BATCH", "ABC123"),
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "LOT", "ABC123"),
        ("https://id.gs1.org/gtin/614141123452/lot/ABC123", "BEST BY", None),
    ],
)
def test_get_element_strings_by_data_title(
    value: str, data_title: str, expected: str | None
) -> None:
    element_string = GS1WebURI.parse(value).element_strings.get(data_title=data_title)

    if expected is None:
        assert element_string is None
    else:
        assert element_string is not None
        assert element_string.value == expected



"""Support for prefixes allocated by GS1.

GS1 prefixes are used to split the allocation space of various number schemes,
e.g. GTIN, GLN, and SSCC.

The allocation space is split twice:

- First, among the national GS1 organizations, represented by
  [`GS1Prefix`][biip.gs1_prefixes.GS1Prefix].
- Secondly, among the companies that apply to their national GS1 organization
  for a number allocation, represented by
  [`GS1CompanyPrefix`][biip.gs1_prefixes.GS1CompanyPrefix].

All prefix data is bundled with Biip. No network access is required.

Examples:
    >>> from biip.gs1_prefixes import GS1CompanyPrefix, GS1Prefix
    >>> GS1Prefix.extract("7044610873466")
    GS1Prefix(value='704', usage='GS1 Norway')
    >>> GS1CompanyPrefix.extract("7044610873466")
    GS1CompanyPrefix(value='704461')
"""

from __future__ import annotations

import json
import lzma
from dataclasses import dataclass
from importlib import resources
from typing import Union

from biip import ParseError

__all__ = [
    "GS1CompanyPrefix",
    "GS1Prefix",
]

_PrefixTrieNode = Union[dict[str, "_PrefixTrieNode"], tuple[int, str]]  # noqa: UP007
_CompanyPrefixTrieNode = Union[dict[str, "_CompanyPrefixTrieNode"], int]  # noqa: UP007


@dataclass(frozen=True)
class GS1Prefix:
    """Prefix assigned by GS1.

    Used to split the allocation space of various number schemes, e.g. GTIN,
    GLN, and SSCC, among GS1 organizations worldwide.

    The GS1 Prefix does not identify the origin of a product, only where the
    number was assigned to a GS1 member organization.

    References:
        https://www.gs1.org/standards/id-keys/company-prefix

    Examples:
        >>> from biip.gs1_prefixes import GS1Prefix
        >>> GS1Prefix.extract("9781492053743")
        GS1Prefix(value='978', usage='Bookland (ISBN)')
    """

    value: str
    """The prefix itself."""

    usage: str
    """Description of who is using the prefix."""

    @classmethod
    def extract(cls, value: str) -> GS1Prefix | None:
        """Extract the GS1 Prefix from the given value.

        Args:
            value: The string to extract a GS1 Prefix from.

        Returns:
            Metadata about the extracted prefix, or `None` if the prefix is unknown.

        Raises:
            ParseError: If the parsing fails.
        """
        if not value.isdecimal():
            msg = f"Failed to get GS1 Prefix from {value!r}."
            raise ParseError(msg)

        node = _GS1_PREFIX_TRIE
        digits = list(value)

        while digits:
            digit = digits.pop(0)

            assert isinstance(node, dict)
            if digit not in node:
                # Prefix is undefined.
                return None
            node = node[digit]

            if isinstance(node, list):
                length, usage = node
                return cls(value=value[:length], usage=usage)

        return None


@dataclass(frozen=True)
class GS1CompanyPrefix:
    """Company prefix assigned by GS1.

    The prefix assigned to a single company.

    Examples:
        >>> from biip.gs1_prefixes import GS1CompanyPrefix
        >>> GS1CompanyPrefix.extract("7044610873466")
        GS1CompanyPrefix(value='704461')
    """

    value: str
    """The company prefix itself."""

    @classmethod
    def extract(cls, value: str) -> GS1CompanyPrefix | None:
        """Extract the GS1 Company Prefix from the given value.

        Args:
            value: The string to extract a GS1 Company Prefix from.
                The value is typically a GLN, GTIN, or an SSCC.

        Returns:
            Metadata about the extracted prefix, or `None` if the prefix is unknown.

        Raises:
            ParseError: If the parsing fails.
        """
        if not value.isdecimal():
            msg = f"Failed to get GS1 Company Prefix from {value!r}."
            raise ParseError(msg)

        node = _GS1_COMPANY_PREFIX_TRIE
        digits = list(value)

        while digits:
            digit = digits.pop(0)

            assert isinstance(node, dict)
            if digit not in node:
                # Prefix is undefined.
                return None
            node = node[digit]

            if isinstance(node, int):
                if node == 0:
                    # Prefix is defined, but has zero length.
                    return None
                return cls(value=value[:node])

        return None


_GS1_PREFIX_TRIE_FILE = (
    resources.files("biip") / "gs1_prefixes" / "_prefix_trie.json.lzma"
)
_GS1_PREFIX_TRIE: _PrefixTrieNode = json.loads(
    lzma.decompress(_GS1_PREFIX_TRIE_FILE.read_bytes()).decode()
)

_GS1_COMPANY_PREFIX_TRIE_FILE = (
    resources.files("biip") / "gs1_prefixes" / "_company_prefix_trie.json.lzma"
)
_GS1_COMPANY_PREFIX_TRIE: _CompanyPrefixTrieNode = json.loads(
    lzma.decompress(_GS1_COMPANY_PREFIX_TRIE_FILE.read_bytes()).decode()
)



r"""Support for GS1 Application Identifiers (AI).

Application identifiers are used to identify the type of data that follows in a
GS1 element string.

All application identifier data is bundled with Biip. No network access is
required.

Examples:
    >>> from biip.gs1_application_identifiers import GS1ApplicationIdentifier
    >>> ai = GS1ApplicationIdentifier.extract("01")
    >>> pprint(ai)
    GS1ApplicationIdentifier(
        ai='01',
        description='Global Trade Item Number (GTIN)',
        data_title='GTIN',
        separator_required=False,
        format='N2+N14'
    )
    >>> ai.pattern
    '^01(\\d{14})$'
"""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from importlib import resources

from biip import ParseError


@dataclass(frozen=True)
class GS1ApplicationIdentifier:
    """GS1 Application Identifier (AI).

    AIs are data field prefixes used in several types of barcodes, including
    GS1-128. The AI defines what semantical meaning and format of the following
    data field.

    AIs standardize how to include e.g. product weight, expiration dates,
    and lot numbers in barcodes.

    References:
        https://www.gs1.org/standards/barcodes/application-identifiers
    """

    ai: str
    """The Application Identifier (AI) itself."""

    description: str
    """Description of the AIs use."""

    data_title: str
    """Commonly used label/abbreviation for the AI."""

    separator_required: bool
    """Whether a separator character is required after element strings of this type.

    This is the case for all AIs that have a variable length.
    """

    format: str
    """Human readable format of the AIs element string."""

    pattern: str = field(repr=False)
    """Regular expression for parsing the AIs element string."""

    @classmethod
    def extract(cls, value: str) -> GS1ApplicationIdentifier:
        """Extract the GS1 Application Identifier (AI) from the given value.

        Args:
            value: The string to extract an AI from.

        Returns:
            Metadata about the extracted AI.

        Raises:
            ParseError: If the parsing fails.

        Examples:
            >>> from biip.gs1_application_identifiers import GS1ApplicationIdentifier
            >>> ai = GS1ApplicationIdentifier.extract("010703206980498815210526100329")
            >>> pprint(ai)
            GS1ApplicationIdentifier(
                ai='01',
                description='Global Trade Item Number (GTIN)',
                data_title='GTIN',
                separator_required=False,
                format='N2+N14'
            )
        """
        for application_identifier in _GS1_APPLICATION_IDENTIFIERS.values():
            if value.startswith(application_identifier.ai):
                return application_identifier

        msg = f"Failed to get GS1 Application Identifier from {value!r}."
        raise ParseError(msg)

    def __len__(self) -> int:
        """Get the length of the Application Identifier code."""
        return len(self.ai)

    def __str__(self) -> str:
        """Get the string representation of the Application Identifier."""
        return f"({self.ai})"


_GS1_APPLICATION_IDENTIFIERS_FILE = (
    resources.files("biip")
    / "gs1_application_identifiers"
    / "_application_identifiers.json"
)
_GS1_APPLICATION_IDENTIFIERS = {
    entry.ai: entry
    for entry in [
        GS1ApplicationIdentifier(**kwargs)
        for kwargs in json.loads(_GS1_APPLICATION_IDENTIFIERS_FILE.read_text())
    ]
}



"""Universal Product Code (UPC).

The `biip.upc` module contains Biip's support for parsing UPC formats.

This class can interpret the following UPC formats:

- UPC-A, 12 digits.
- UPC-E, 6 digits, with implicit number system 0 and no check digit.
- UPC-E, 7 digits, with explicit number system and no check digit.
- UPC-E, 8 digits, with explicit number system and a check digit.

If you only want to parse UPCs, you can import the UPC parser directly
instead of using [`biip.parse()`][biip.parse].

    >>> from biip.upc import Upc

If parsing succeeds, it returns a [`Upc`][biip.upc.Upc] object.

    >>> upc_a = Upc.parse("042100005264")
    >>> pprint(upc_a)
    Upc(
        value='042100005264',
        format=UpcFormat.UPC_A,
        number_system_digit=0,
        payload='04210000526',
        check_digit=4
    )

A subset of the UPC-A values can be converted to a shorter UPC-E format by
suppressing zeros using [`as_upc_e()`][biip.upc.Upc.as_upc_e].

    >>> upc_a.as_upc_e()
    '04252614'

All UPC-E values can be expanded to an UPC-A using
[`as_upc_a()`][biip.upc.Upc.as_upc_a].

    >>> upc_e = Upc.parse("04252614")
    >>> pprint(upc_e)
    Upc(
        value='04252614',
        format=UpcFormat.UPC_E,
        number_system_digit=0,
        payload='0425261',
        check_digit=4
    )
    >>> upc_e.as_upc_a()
    '042100005264'

UPC is a subset of the later GTIN standard: An UPC-A value is also a valid
GTIN-12 value.

    >>> upc_e.as_gtin_12()
    '042100005264'

The canonical format for persisting UPCs to e.g. a database is GTIN-14.

    >>> upc_e.as_gtin_14()
    '00042100005264'
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from biip import EncodeError, ParseConfig, ParseError
from biip.checksums import gs1_standard_check_digit


class UpcFormat(Enum):
    """Enum of UPC formats."""

    UPC_A = "upc_a"
    """UPC-A"""

    UPC_E = "upc_e"
    """UPC-E"""

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"UpcFormat.{self.name}"


@dataclass(frozen=True)
class Upc:
    """Data class containing an UPC."""

    value: str
    """Raw unprocessed value."""

    format: UpcFormat
    """UPC format, either UPC-A or UPC-E."""

    number_system_digit: int
    """Number system digit."""

    payload: str
    """The actual payload.

    Including number system digit, manufacturer code, and product code. Excludes
    the check digit.
    """

    check_digit: int | None = None
    """Check digit used to check if the UPC-A as a whole is valid.

    Set for UPC-A, but not set for UPC-E.
    """

    @classmethod
    def parse(
        cls,
        value: str,
        *,
        config: ParseConfig | None = None,  # noqa: ARG003
    ) -> Upc:
        """Parse the given value into a [`Upc`][biip.upc.Upc] object.

        The checksum is guaranteed to be valid if an UPC object is returned.

        Args:
            value: The value to parse.
            config: Configuration options for parsing.

        Returns:
            UPC data structure with the successfully extracted data.

        Raises:
            ParseError: If the parsing fails.
        """
        value = value.strip()

        length = len(value)
        if length not in (6, 7, 8, 12):
            msg = (
                f"Failed to parse {value!r} as UPC: "
                f"Expected 6, 7, 8, or 12 digits, got {length}."
            )
            raise ParseError(msg)

        if not value.isdecimal():
            msg = f"Failed to parse {value!r} as UPC: Expected a numerical value."
            raise ParseError(msg)

        if length == 12:
            return cls._parse_upc_a(value)

        if length in (6, 7, 8):
            return cls._parse_upc_e(value)

        msg = "Unhandled UPC length. This is a bug."  # pragma: no cover
        raise NotImplementedError(msg)  # pragma: no cover

    @classmethod
    def _parse_upc_a(cls, value: str) -> Upc:
        assert len(value) == 12

        payload = value[:-1]
        number_system_digit = int(value[0])
        check_digit = int(value[-1])

        calculated_check_digit = gs1_standard_check_digit(payload)
        if check_digit != calculated_check_digit:
            msg = (
                f"Invalid UPC-A check digit for {value!r}: "
                f"Expected {calculated_check_digit!r}, got {check_digit!r}."
            )
            raise ParseError(msg)

        return cls(
            value=value,
            format=UpcFormat.UPC_A,
            payload=payload,
            number_system_digit=number_system_digit,
            check_digit=check_digit,
        )

    @classmethod
    def _parse_upc_e(cls, value: str) -> Upc:
        length = len(value)
        assert length in (6, 7, 8)

        if length == 6:
            # Implicit number system 0, no check digit.
            number_system_digit = 0
            payload = f"{number_system_digit}{value}"
            upc_a_payload = _upc_e_to_upc_a_expansion(f"{payload}0")[:-1]
            check_digit = gs1_standard_check_digit(upc_a_payload)
        elif length == 7:
            # Explicit number system, no check digit.
            number_system_digit = int(value[0])
            payload = value
            upc_a_payload = _upc_e_to_upc_a_expansion(f"{payload}0")[:-1]
            check_digit = gs1_standard_check_digit(upc_a_payload)
        elif length == 8:
            # Explicit number system and check digit.
            number_system_digit = int(value[0])
            payload = value[:-1]
            check_digit = int(value[-1])
        else:
            msg = "Unhandled UPC-E length. This is a bug."  # pragma: no cover
            raise Exception(msg)  # noqa: TRY002  # pragma: no cover

        # Control that the number system digit is correct.
        if number_system_digit not in (0, 1):
            msg = (
                f"Invalid UPC-E number system for {value!r}: "
                f"Expected 0 or 1, got {number_system_digit!r}."
            )
            raise ParseError(msg)

        # Control that check digit is correct.
        upc_a_payload = _upc_e_to_upc_a_expansion(f"{payload}{check_digit}")[:-1]
        calculated_check_digit = gs1_standard_check_digit(upc_a_payload)
        if check_digit != calculated_check_digit:
            msg = (
                f"Invalid UPC-E check digit for {value!r}: "
                f"Expected {calculated_check_digit!r}, got {check_digit!r}."
            )
            raise ParseError(msg)

        return cls(
            value=value,
            format=UpcFormat.UPC_E,
            payload=payload,
            number_system_digit=number_system_digit,
            check_digit=check_digit,
        )

    def as_upc_a(self) -> str:
        """Format as UPC-A.

        Returns:
            A string with the UPC encoded as UPC-A.

        References:
            GS1 General Specifications, section 5.2.2.4.2
        """
        if self.format == UpcFormat.UPC_A:
            return f"{self.payload}{self.check_digit}"

        if self.format == UpcFormat.UPC_E:
            return _upc_e_to_upc_a_expansion(f"{self.payload}{self.check_digit}")

        msg = (  # pragma: no cover
            "Unhandled case while formatting as UPC-A. This is a bug."
        )
        raise NotImplementedError(msg)  # pragma: no cover

    def as_upc_e(self) -> str:
        """Format as UPC-E.

        Returns:
            A string with the UPC encoded as UPC-E, if possible.

        Raises:
            EncodeError: If encoding as UPC-E fails.

        References:
            GS1 General Specifications, section 5.2.2.4.1
        """
        if self.format == UpcFormat.UPC_A:
            return _upc_a_to_upc_e_suppression(f"{self.payload}{self.check_digit}")

        if self.format == UpcFormat.UPC_E:
            return f"{self.payload}{self.check_digit}"

        msg = (  # pragma: no cover
            "Unhandled case while formatting as UPC-E. This is a bug."
        )
        raise NotImplementedError(msg)  # pragma: no cover

    def as_gtin_12(self) -> str:
        """Format as GTIN-12."""
        from biip.gtin import Gtin

        return Gtin.parse(self.as_upc_a()).as_gtin_12()

    def as_gtin_13(self) -> str:
        """Format as GTIN-13."""
        from biip.gtin import Gtin

        return Gtin.parse(self.as_upc_a()).as_gtin_13()

    def as_gtin_14(self) -> str:
        """Format as GTIN-14."""
        from biip.gtin import Gtin

        return Gtin.parse(self.as_upc_a()).as_gtin_14()


def _upc_e_to_upc_a_expansion(value: str) -> str:
    assert len(value) == 8
    assert value.isdecimal()

    last_digit = int(value[6])
    check_digit = int(value[7])

    if last_digit in (0, 1, 2):
        return f"{value[:3]}{last_digit}0000{value[3:6]}{check_digit}"

    if last_digit == 3:
        return f"{value[:4]}00000{value[4:6]}{check_digit}"

    if last_digit == 4:
        return f"{value[:5]}00000{value[5]}{check_digit}"

    if last_digit in (5, 6, 7, 8, 9):
        return f"{value[:6]}0000{last_digit}{check_digit}"

    msg = (  # pragma: no cover
        "Unhandled case while expanding UPC-E to UPC-A. This is a bug."
    )
    raise Exception(msg)  # noqa: TRY002  # pragma: no cover


def _upc_a_to_upc_e_suppression(value: str) -> str:
    assert len(value) == 12
    assert value.isdecimal()

    check_digit = int(value[11])

    if int(value[10]) in (5, 6, 7, 8, 9) and value[6:10] == "0000" and value[5] != "0":
        # UPC-E suppression, condition A
        return f"{value[:6]}{value[10]}{check_digit}"

    if value[5:10] == "00000" and value[4] != "0":
        # UPC-E suppression, condition B
        return f"{value[:5]}{value[10]}4{check_digit}"

    if value[4:8] == "0000" and int(value[3]) in (0, 1, 2):
        # UPC-E suppression, condition C
        return f"{value[:3]}{value[8:11]}{value[3]}{check_digit}"

    if value[4:9] == "00000" and int(value[3]) in range(3, 10):
        # UPC-E suppression, condition D
        return f"{value[:4]}{value[9:11]}3{check_digit}"

    msg = f"UPC-A {value!r} cannot be represented as UPC-E."
    raise EncodeError(msg)



"""Support for Symbology Identifiers.

Symbology Identifiers is a standardized way to identify what type of barcode
symbology was used to encode the following data.

The Symbology Identifiers are a few extra characters that may be prefixed to
the scanned data by the barcode scanning hardware. The software interpreting
the barcode may use the Symbology Identifier to differentiate how to handle
the barcode, but must at the very least be able to strip and ignore the extra
characters.

Examples:
    >>> from biip.symbology import SymbologyIdentifier
    >>> si = SymbologyIdentifier.extract("]E05901234123457")
    >>> pprint(si)
    SymbologyIdentifier(
        value=']E0',
        iso_symbology=ISOSymbology.EAN_UPC,
        modifiers='0',
        gs1_symbology=GS1Symbology.EAN_13
    )
    >>> si = SymbologyIdentifier.extract("]I198765432109213")
    >>> pprint(si)
    SymbologyIdentifier(
        value=']I1',
        iso_symbology=ISOSymbology.ITF,
        modifiers='1',
        gs1_symbology=GS1Symbology.ITF_14
    )

References:
    ISO/IEC 15424:2008.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from biip import ParseError

__all__ = [
    "GS1Symbology",
    "ISOSymbology",
    "SymbologyIdentifier",
]


class ISOSymbology(Enum):
    """Enum of barcode symbologies that are supported by Symbology Identifers.

    References:
        ISO/IEC 15424:2008, table 1.
    """

    CODE_39 = "A"
    """Code 39"""

    TELEPEN = "B"
    """Telepen"""

    CODE_128 = "C"
    """Code 128"""

    CODE_ONE = "D"
    """Code One"""

    EAN_UPC = "E"
    """EAN/UPC"""

    CODABAR = "F"
    """Codabar"""

    CODE_93 = "G"
    """Code 93"""

    CODE_11 = "H"
    """Code 11"""

    ITF = "I"
    """ITF (Interleaved 2 of 5)"""

    CODE_16K = "K"
    """Code 16K"""

    PDF417 = "L"
    """PDF417 and MicroPDF417"""

    MSI = "M"
    """MSI"""

    ANKER = "N"
    """Anker"""

    CODABLOCK = "O"
    """Codablock"""

    PLESSEY_CODE = "P"
    """Plessey Code"""

    QR_CODE = "Q"
    """QR Code and QR Code 2005"""

    STRAIGHT_2_OF_5_WITH_2_BAR_START_STOP_CODE = "R"
    """Straigt 2 of 5 (with two bar start/stop codes)"""

    STRAIGHT_2_OF_5_WITH_3_BAR_START_STOP_CODE = "S"
    """Straigt 2 of 5 (with three bar start/stop codes)"""

    CODE_49 = "T"
    """Code 49"""

    MAXICODE = "U"
    """MaxiCode"""

    OTHER_BARCODE = "X"
    """Other barcode"""

    SYSTEM_EXPANSION = "Y"
    """System expansion"""

    NON_BARCODE = "Z"
    """Non-barcode"""

    CHANNEL_CODE = "c"
    """Channel Code"""

    DATA_MATRIX = "d"
    """Data Matrix"""

    RSS_EAN_UCC_COMPOSITE = "e"
    """RSS and EAN.UCC Composite"""

    OCR = "o"
    """OCR (Optical Character Recognition)"""

    POSICODE = "p"
    """PosiCode"""

    SUPERCODE = "s"
    """SuperCode"""

    AZTEC_CODE = "z"
    """Aztec Code"""

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"ISOSymbology.{self.name}"


class GS1Symbology(Enum):
    """Enum of Symbology Identifiers used in the GS1 system.

    References:
        - GS1 General Specifications, figure 5.1.2-2.
        - ISO/IEC 15424:2008.
    """

    EAN_13 = "E0"
    """EAN-13, UPC-A, or UPC-E."""

    EAN_TWO_DIGIT_ADD_ON = "E1"
    """Two-digit add-on symbol for EAN-13."""

    EAN_FIVE_DIGIT_ADD_ON = "E2"
    """Five-digit add-on symbol for EAN-13."""

    EAN_13_WITH_ADD_ON = "E3"
    """EAN-13, UPC-A, or UPC-E with add-on symbol."""

    EAN_8 = "E4"
    """EAN-8"""

    ITF_14 = "I1"
    """ITF-14"""

    GS1_128 = "C1"
    """GS1-128"""

    GS1_DATABAR = "e0"
    """GS1 DataBar"""

    GS1_COMPOSITE_WITH_SEPARATOR_CHAR = "e1"
    """GS1 Composite. Data packet follows an encoded symbol separator character."""

    GS1_COMPOSITE_WITH_ESCAPE_CHAR = "e2"
    """GS1 Composite. Data packet follows an escape mechanism character."""

    GS1_DATAMATRIX = "d2"
    """GS1 DataMatrix"""

    GS1_QR_CODE = "Q3"
    """GS1 QR Code"""

    GS1_DOTCODE = "J1"
    """GS1 DotCode"""

    @classmethod
    def with_gs1_messages(cls) -> set[GS1Symbology]:
        """Symbologies that may contain GS1 messages."""
        return {
            cls.GS1_128,
            cls.GS1_DATABAR,
            cls.GS1_DATAMATRIX,
            cls.GS1_QR_CODE,
            cls.GS1_DOTCODE,
        }

    @classmethod
    def with_gs1_web_uri(cls) -> set[GS1Symbology]:
        """Symbologies that may contain GS1 Web URIs."""
        return {cls.GS1_DATAMATRIX, cls.GS1_QR_CODE}

    @classmethod
    def with_gtin(cls) -> set[GS1Symbology]:
        """Symbologies that may contain GTINs."""
        return {cls.EAN_13, cls.EAN_13_WITH_ADD_ON, cls.EAN_8, cls.ITF_14}

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"GS1Symbology.{self.name}"


@dataclass(frozen=True)
class SymbologyIdentifier:
    """Data class containing a Symbology Identifier."""

    value: str
    """Raw unprocessed value."""

    iso_symbology: ISOSymbology
    """The recognized ISO symbology."""

    modifiers: str
    """Symbology modifiers.

    Refer to [`gs1_symbology`][biip.symbology.SymbologyIdentifier.gs1_symbology]
    or ISO/IEC 15424 for interpretation.
    """

    gs1_symbology: GS1Symbology | None = None
    """If the Symbology Identifier is used in the GS1 system,
    this field is set to indicate how to interpret the following data.
    """

    @classmethod
    def extract(cls, value: str) -> SymbologyIdentifier:
        """Extract the Symbology Identifier from the given value.

        Args:
            value: The string to extract a Symbology Identifier from.

        Returns:
            Metadata about the extracted Symbology Identifier.

        Raises:
            ParseError: If the parsing fails.
        """
        if not value.startswith("]"):
            msg = (
                f"Failed to get Symbology Identifier from {value!r}. "
                "No initial ']' flag character found."
            )
            raise ParseError(msg)

        try:
            iso_symbology = ISOSymbology(value[1])
        except ValueError as exc:
            msg = (
                f"Failed to get Symbology Identifier from {value!r}. "
                f"{value[1]!r} is not a recognized code character."
            )
            raise ParseError(msg) from exc

        if iso_symbology == ISOSymbology.SYSTEM_EXPANSION:
            modifiers_length = int(value[2]) + 1
        else:
            modifiers_length = 1

        modifiers = value[2 : 2 + modifiers_length]

        value = f"]{iso_symbology.value}{modifiers}"

        gs1_symbology: GS1Symbology | None
        try:
            gs1_symbology = GS1Symbology(f"{iso_symbology.value}{modifiers}")
        except ValueError:
            gs1_symbology = None

        return cls(
            value=value,
            iso_symbology=iso_symbology,
            modifiers=modifiers,
            gs1_symbology=gs1_symbology,
        )

    def __len__(self) -> int:
        """Get the length of the Symbology Identfier."""
        return len(self.value)

    def __str__(self) -> str:
        """Get the string representation of the Symbology Identifier."""
        return self.value



"""Test setup for doctests."""

# For the fixture to be available when running doctests, this file must be placed
# in the `src/` tree or outside of it, not in the `tests/` tree.

from typing import Any

import pytest
from rich.pretty import pprint


def doctest_print(*args: Any) -> None:  # noqa: ANN401, D103
    return pprint(
        *args,
        expand_all=True,
        indent_guides=False,
    )


@pytest.fixture(autouse=True, scope="session")
def pprint_setup(doctest_namespace: dict[str, Any]) -> None:  # noqa: D103
    doctest_namespace["pprint"] = doctest_print
